from .lutils.bitboard import iterBits
from .lutils.LBoard import LBoard
from .lutils.lmove import RANK, FILE, FCORD, FLAG, PROMOTE_PIECE
from .Piece import Piece
from .Cord import Cord
from .const import (
    A1,
    A8,
    B1,
    B8,
    C1,
    C8,
    D1,
    D8,
    E1,
    E8,
    F1,
    F8,
    G1,
    G8,
    H1,
    H8,
    BISHOP,
    ROOK,
    ROOK_PROMOTION,
    QUEEN_PROMOTION,
    KNIGHT_PROMOTION,
    BLACK,
    FEN_START,
    WHITE,
    NORMALCHESS,
    PAWN,
    BISHOP_PROMOTION,
    KNIGHT,
    QUEEN,
    KING,
    DROP_VARIANTS,
    DROP,
    ATOMICCHESS,
    ENPASSANT,
    FISCHERRANDOMCHESS,
    QUEEN_CASTLE,
    CRAZYHOUSECHESS,
    KING_CASTLE,
    WILDCASTLECHESS,
    PROMOTIONS,
    WILDCASTLESHUFFLECHESS,
    FAN_PIECES,
    GATINGS,
    HAWK,
    HAWK_GATE,
    ELEPHANT,
    HAWK_GATE_AT_ROOK,
    ELEPHANT_GATE_AT_ROOK,
    SCHESS,
)


def reverse_enum(L):
    for index in reversed(range(len(L))):
        yield index, L[index]


class Board:
    """Board is a thin layer above LBoard, adding the Piece objects, which are
    needed for animation in BoardView.
    In contrast to LBoard, Board is immutable, which means it will clone
    itself each time you apply a move to it.
    Caveat: As the only objects, the Piece objects in the self.data lists
    will not be cloned, to make animation state preserve between moves"""

    variant = NORMALCHESS
    RANKS = 8
    FILES = 8
    HOLDING_FILES = ((FILES + 3, FILES + 2, FILES + 1), (-4, -3, -2))
    PROMOTION_ZONE = (
        (A8, B8, C8, D8, E8, F8, G8, H8),
        (A1, B1, C1, D1, E1, F1, G1, H1),
    )
    PROMOTIONS = (QUEEN_PROMOTION, ROOK_PROMOTION, BISHOP_PROMOTION, KNIGHT_PROMOTION)

    def __init__(self, setup=False, lboard=None):
        self.data = [dict(enumerate([None] * self.FILES)) for i in range(self.RANKS)]
        if lboard is None:
            self.board = LBoard(self.variant)
        else:
            self.board = lboard
        self.board.pieceBoard = self

        # Set True in interactive lesson games after they happened
        self.played = False

        if setup:
            if lboard is None:
                if setup is True:
                    self.board.applyFen(FEN_START)
                elif isinstance(setup, str):
                    self.board.applyFen(setup)

            for color in (BLACK, WHITE):
                pieces = self.board.boards[color]
                for piece in (PAWN, KNIGHT, BISHOP, ROOK, QUEEN, KING, HAWK, ELEPHANT):
                    for cord in iterBits(pieces[piece]):
                        self.data[RANK(cord)][FILE(cord)] = Piece(color, piece)

            if self.variant in DROP_VARIANTS:
                for color in (BLACK, WHITE):
                    holding = self.board.holding[color]
                    for piece in holding:
                        for i in range(holding[piece]):
                            self[self.newHoldingCord(color, 1)] = Piece(color, piece)

    def getHoldingCord(self, color, piece, piece_color=None):
        """Get the chord of first occurrence of piece in given color holding"""

        enum = reverse_enum if color == WHITE else enumerate
        for x_loc in self.HOLDING_FILES[color]:
            for y_loc, row in enum(self.data):
                if (row.get(x_loc) is not None) and row.get(x_loc).piece == piece:
                    # In S-chess holding may contain our un-gated and captured H/E at the same time
                    # but in case of gating we are looking for our own HAWK/ELEPHANT only!
                    if piece_color is not None and row.get(x_loc).color != piece_color:
                        continue
                    return Cord(x_loc, y_loc)

    def newHoldingCord(self, color, nth=1):
        """Find the nth empty slot in given color holding.
        In atomic explosions nth can be > 1.
        """

        enum = reverse_enum if color == BLACK else enumerate
        empty = 0
        for x_loc in reversed(self.HOLDING_FILES[color]):
            for y_loc, row in enum(self.data):
                if row.get(x_loc) is None:
                    empty += 1
                    if empty == nth:
                        return Cord(x_loc, y_loc)

    def getHoldingPieces(self, color):
        """Get the list of pieces from given color holding"""
        pieces = []
        for x_loc in self.HOLDING_FILES[color]:
            for row in self.data:
                if row.get(x_loc) is not None:
                    pieces.append(row.get(x_loc))
        return pieces

    def popPieceFromHolding(self, color, piece):
        """Remove and return a piece in given color holding"""

        for x_loc in self.HOLDING_FILES[color]:
            for row in self.data:
                if (row.get(x_loc) is not None) and row.get(x_loc).piece == piece:
                    piece = row.get(x_loc)
                    del row[x_loc]
                    return piece
        return None

    def reorderHolding(self, color):
        """Reorder captured pieces by their value"""
        pieces = []
        for piece in (PAWN, KNIGHT, BISHOP, ROOK, QUEEN, KING):
            while True:
                col_piece = self.popPieceFromHolding(color, piece)
                if col_piece is not None:
                    pieces.append(col_piece)
                else:
                    break
        for piece in pieces:
            self[self.newHoldingCord(color, 1)] = piece

    def simulateMove(self, board1, move):
        moved = []
        new = []
        dead = []

        cord0, cord1 = move.cords
        flag = FLAG(move.move)

        # null move or SITTUYINCHESS in place promotion
        if cord0 == cord1 and flag != DROP and flag != QUEEN_PROMOTION:
            return moved, new, dead

        if flag == DROP:
            piece = FCORD(move.move)
            cord0 = self.getHoldingCord(self.color, piece)
            moved.append((self[cord0], cord0))
            # add all captured pieces to "new" list to enforce repainting them after a possible reordering
            new = board1.getHoldingPieces(self.color)
            dead = new
            return moved, new, dead

        kcastle = flag == KING_CASTLE or (
            self.variant == SCHESS
            and (
                (self[move.cord0].piece == KING and cord0.x - cord1.x == -2)
                or (
                    flag in (HAWK_GATE_AT_ROOK, ELEPHANT_GATE_AT_ROOK)
                    and cord0.x - cord1.x > 0
                )
            )
        )
        qcastle = flag == QUEEN_CASTLE or (
            self.variant == SCHESS
            and (
                (self[move.cord0].piece == KING and cord0.x - cord1.x == 2)
                or (
                    flag in (HAWK_GATE_AT_ROOK, ELEPHANT_GATE_AT_ROOK)
                    and cord0.x - cord1.x < 0
                )
            )
        )

        if self.variant == ATOMICCHESS and (self[cord1] or flag == ENPASSANT):
            # Sequence nubers of next newHoldingCord of WHITE and BLACK
            nth = [0, 0]

            piece = self[cord0]
            nth[1 - piece.color] += 1
            cord = self.newHoldingCord(1 - piece.color, nth[1 - piece.color])
            moved.append((board1[cord], cord0))
            new.append(board1[cord])
        else:
            if flag in PROMOTIONS:
                dead.append(self[cord0])
            else:
                moved.append((self[cord0], cord0))

        # capture
        if self[cord1] and not (
            flag
            in (QUEEN_CASTLE, KING_CASTLE, HAWK_GATE_AT_ROOK, ELEPHANT_GATE_AT_ROOK)
        ):
            piece = (
                PAWN
                if self.variant == CRAZYHOUSECHESS and self[cord1].promoted
                else self[cord1].piece
            )
            cord = board1.getHoldingCord(self.color, piece)
            moved.append((board1[cord], cord1))
            # add all captured pieces to "new" list to enforce repainting them after a possible reordering
            new = board1.getHoldingPieces(self.color)

            if self.variant == ATOMICCHESS:
                nth[self.color] += 1
                from pychess.Variants.atomic import cordsAround

                for acord in cordsAround(cord1):
                    piece = self[acord]
                    if piece and piece.piece != PAWN and acord != cord0:
                        nth[1 - piece.color] += 1
                        cord = self.newHoldingCord(
                            1 - piece.color, nth[1 - piece.color]
                        )
                        moved.append((board1[cord], acord))
                        new.append(board1[cord])

        if kcastle or qcastle:
            side = 0 if qcastle else 1
            if FILE(cord0.x) == 3 and self.board.variant in (
                WILDCASTLECHESS,
                WILDCASTLESHUFFLECHESS,
            ):
                side = 0 if side == 1 else 1
            rook = self.board.ini_rooks[self.color][side]
            moved.append((self[Cord(rook)], Cord(rook)))

        elif flag in PROMOTIONS:
            newPiece = board1[cord1]
            moved.append((newPiece, cord0))
            new.append(newPiece)

        elif flag == ENPASSANT:
            shift = -1 if self.color == WHITE else 1
            ep_cord = Cord(cord1.x, cord1.y + shift)
            moved.append((self[ep_cord], ep_cord))
            # add all captured pieces to "new" list to enforce repainting them after a possible reordering
            new = board1.getHoldingPieces(self.color)

        if flag in GATINGS:
            gpiece = HAWK if flag in (HAWK_GATE, HAWK_GATE_AT_ROOK) else ELEPHANT
            cord0 = self.getHoldingCord(self.color, gpiece)
            moved.append((self[cord0], cord0))
            # add all captured pieces to "new" list to enforce repainting them after a possible reordering
            new = board1.getHoldingPieces(self.color)

        return moved, new, dead

    def simulateUnmove(self, board1, move):
        moved = []
        new = []
        dead = []

        cord0, cord1 = move.cords
        flag = FLAG(move.move)

        # null move or SITTUYINCHESS in place promotion
        if cord0 == cord1 and flag != DROP and flag != QUEEN_PROMOTION:
            return moved, new, dead

        if self.variant == ATOMICCHESS and (board1[cord1] or flag == ENPASSANT):
            piece = board1[cord0].piece
            cord = self.getHoldingCord(self.color, piece)
            moved.append((self[cord], cord))
            self[cord].opacity = 1
            dead.append(self[cord])
        elif not (
            self.variant == FISCHERRANDOMCHESS and flag in (QUEEN_CASTLE, KING_CASTLE)
        ):
            moved.append((self[cord1], cord1))

        kcastle = flag == KING_CASTLE or (
            self.variant == SCHESS
            and (
                (
                    self[move.cord1] is not None
                    and self[move.cord1].piece == KING
                    and cord0.x - cord1.x == -2
                )
                or (
                    flag in (HAWK_GATE_AT_ROOK, ELEPHANT_GATE_AT_ROOK)
                    and cord0.x - cord1.x > 0
                )
            )
        )
        qcastle = flag == QUEEN_CASTLE or (
            self.variant == SCHESS
            and (
                (
                    self[move.cord1] is not None
                    and self[move.cord1].piece == KING
                    and cord0.x - cord1.x == 2
                )
                or (
                    flag in (HAWK_GATE_AT_ROOK, ELEPHANT_GATE_AT_ROOK)
                    and cord0.x - cord1.x < 0
                )
            )
        )

        # capture
        if board1[cord1] and not (
            flag
            in (QUEEN_CASTLE, KING_CASTLE, HAWK_GATE_AT_ROOK, ELEPHANT_GATE_AT_ROOK)
        ):
            piece = (
                PAWN
                if self.variant == CRAZYHOUSECHESS and board1[cord1].promoted
                else board1[cord1].piece
            )
            cord = self.getHoldingCord(1 - self.color, piece)
            moved.append((self[cord], cord))
            self[cord].opacity = 1
            # add all captured pieces to "new" list to enforce repainting them after a possible reordering
            new = self.getHoldingPieces(self.color)
            dead.append(self[cord])

            if self.variant == ATOMICCHESS:
                from pychess.Variants.atomic import cordsAround

                for acord in cordsAround(cord1):
                    piece = board1[acord]
                    if piece and piece.piece != PAWN and acord != cord0:
                        piece.opacity = 0
                        cord = self.getHoldingCord(1 - piece.color, piece.piece)
                        moved.append((self[cord], cord))
                        self[cord].opacity = 1
                        dead.append(self[cord])

        if kcastle or qcastle:
            side = 0 if qcastle else 1
            if FILE(cord0.x) == 3 and self.board.variant in (
                WILDCASTLECHESS,
                WILDCASTLESHUFFLECHESS,
            ):
                side = 0 if side == 1 else 1
            rook = self.board.fin_rooks[board1.color][side]
            moved.append((self[Cord(rook)], Cord(rook)))

        elif flag in PROMOTIONS:
            newPiece = board1[cord0]
            moved.append((newPiece, cord1))
            new.append(newPiece)

        elif flag == ENPASSANT:
            cord = self.getHoldingCord(1 - self.color, PAWN)
            moved.append((self[cord], cord))
            self[cord].opacity = 1
            # add all captured pieces to "new" list to enforce repainting them after a possible reordering
            new = self.getHoldingPieces(self.color)
            dead.append(self[cord])

        elif flag in GATINGS:
            # add all gated pieces to "new" list to enforce repainting them after a possible reordering
            new = self.getHoldingPieces(1 - self.color)

        return moved, new, dead

    def move(self, move, lboard=None):
        """Creates a new Board object cloning itself then applying
        the move.move to the clone Board's lboard.
        If lboard param was given, it will be used when cloning,
        and move will not be applyed, just the high level Piece
        objects will be adjusted."""

        # Sequence nubers of next newHoldingCord of WHITE and BLACK
        nth = [0, 0]

        flag = FLAG(move.move)
        if flag != DROP:
            assert self[move.cord0], f"{move} {self.asFen()}"

        newBoard = self.clone(lboard=lboard)
        if lboard is None:
            newBoard.board.applyMove(move.move)
        cord0, cord1 = move.cords

        kcastle = flag == KING_CASTLE or (
            self.variant == SCHESS
            and (
                (self[move.cord0].piece == KING and cord0.x - cord1.x == -2)
                or (
                    flag in (HAWK_GATE_AT_ROOK, ELEPHANT_GATE_AT_ROOK)
                    and cord0.x - cord1.x > 0
                )
            )
        )
        qcastle = flag == QUEEN_CASTLE or (
            self.variant == SCHESS
            and (
                (self[move.cord0].piece == KING and cord0.x - cord1.x == 2)
                or (
                    flag in (HAWK_GATE_AT_ROOK, ELEPHANT_GATE_AT_ROOK)
                    and cord0.x - cord1.x < 0
                )
            )
        )

        if kcastle or qcastle:
            gcord = cord0  # save gating cord

        # capture
        if (
            (self[move.cord1] is not None or flag == ENPASSANT)
            and not (cord0 == cord1 and flag != DROP and flag != QUEEN_PROMOTION)
            and not (kcastle or qcastle)
        ):
            if self.variant == CRAZYHOUSECHESS:
                piece = (
                    PAWN
                    if flag == ENPASSANT or self[move.cord1].promoted
                    else self[move.cord1].piece
                )
                new_piece = Piece(self.color, piece, captured=True)
            else:
                piece = PAWN if flag == ENPASSANT else self[move.cord1].piece
                new_piece = Piece(1 - self.color, piece, captured=True)
            nth[self.color] += 1
            newBoard[self.newHoldingCord(self.color, nth[self.color])] = new_piece

            if self.variant == ATOMICCHESS:
                from pychess.Variants.atomic import cordsAround

                for acord in cordsAround(move.cord1):
                    piece = self[acord]
                    if piece and piece.piece != PAWN and acord != cord0:
                        new_piece = Piece(piece.color, piece.piece, captured=True)
                        nth[1 - piece.color] += 1
                        newBoard[
                            self.newHoldingCord(1 - piece.color, nth[1 - piece.color])
                        ] = new_piece
                        newBoard[acord] = None

        if flag == DROP:
            piece = FCORD(move.move)
            holding_coord = self.getHoldingCord(self.color, piece)
            if holding_coord is None:
                newBoard[cord1] = Piece(self.color, piece)
            else:
                newBoard[cord1] = newBoard[holding_coord]
                newBoard[cord1].captured = False
                newBoard[holding_coord] = None
        else:
            if (
                self.variant == ATOMICCHESS
                and (flag == ENPASSANT or self[move.cord1] is not None)
                and (flag not in (QUEEN_CASTLE, KING_CASTLE))
            ):
                piece = self[move.cord0].piece
                new_piece = Piece(self.color, piece, captured=True)
                nth[1 - self.color] += 1
                newBoard[self.newHoldingCord(1 - self.color, nth[1 - self.color])] = (
                    new_piece
                )
                newBoard[cord1] = None
            else:
                if kcastle or qcastle:
                    if flag in (HAWK_GATE_AT_ROOK, ELEPHANT_GATE_AT_ROOK):
                        king = newBoard[cord1]
                    else:
                        king = newBoard[cord0]
                else:
                    newBoard[cord1] = newBoard[cord0]

        if (
            cord0 != cord1
            and flag != DROP
            and flag != HAWK_GATE_AT_ROOK
            and flag != ELEPHANT_GATE_AT_ROOK
        ):
            newBoard[cord0] = None

        if kcastle or qcastle:
            side = 0 if qcastle else 1
            if FILE(cord0.x) == 3 and self.board.variant in (
                WILDCASTLECHESS,
                WILDCASTLESHUFFLECHESS,
            ):
                side = 0 if side == 1 else 1
            inirook = self.board.ini_rooks[self.color][side]
            finrook = self.board.fin_rooks[self.color][side]
            newBoard[Cord(finrook)] = newBoard[Cord(inirook)]
            if inirook != finrook:
                newBoard[Cord(inirook)] = None
            if flag in (HAWK_GATE_AT_ROOK, ELEPHANT_GATE_AT_ROOK):
                newBoard[cord1] = None
            finking = self.board.fin_kings[self.color][side]
            newBoard[Cord(finking)] = king

        if flag in GATINGS:
            gpiece = HAWK if flag in (HAWK_GATE, HAWK_GATE_AT_ROOK) else ELEPHANT
            holding_coord = self.getHoldingCord(
                self.color, gpiece, piece_color=self.color
            )
            newBoard[gcord if (kcastle or qcastle) else cord0] = newBoard[holding_coord]
            newBoard[holding_coord] = None

        if flag in PROMOTIONS:
            new_piece = Piece(self.color, PROMOTE_PIECE(flag))
            new_piece.promoted = True
            newBoard[cord1] = new_piece

        elif flag == ENPASSANT:
            newBoard[Cord(cord1.x, cord0.y)] = None

        if flag == DROP or flag == ENPASSANT or self[move.cord1] is not None:
            newBoard.reorderHolding(self.color)

        return newBoard

    def switchColor(self):
        """Switches the current color to move and unsets the enpassant cord.
        Mostly to be used by inversed analyzers"""
        new_board = self.setColor(1 - self.color)
        new_board.board.next = self.board.next
        return new_board

    def _get_enpassant(self):
        if self.board.enpassant is not None:
            return Cord(self.board.enpassant)
        return None

    enpassant = property(_get_enpassant)

    def setColor(self, color):
        newBoard = self.clone()
        newBoard.board.setColor(color)
        newBoard.board.setEnpassant(None)
        return newBoard

    def _get_color(self):
        return self.board.color

    color = property(_get_color)

    def _get_ply(self):
        return self.board.plyCount

    ply = property(_get_ply)

    def asFen(self, enable_bfen=True):
        return self.board.asFen(enable_bfen)

    def __repr__(self):
        return repr(self.board)

    def __getitem__(self, cord):
        return self.data[cord.y].get(cord.x)

    def __setitem__(self, cord, piece):
        self.data[cord.y][cord.x] = piece

    def clone(self, lboard=None):
        if lboard is None:
            lboard = self.board.clone()

        if self.variant != NORMALCHESS:
            from pychess.Variants import variants

            newBoard = variants[self.variant]()
        else:
            newBoard = Board()
        newBoard.board = lboard
        newBoard.board.pieceBoard = newBoard

        for y, row in enumerate(self.data):
            for x, piece in row.items():
                newBoard.data[y][x] = piece

        return newBoard

    def __eq__(self, other):
        if not isinstance(self, type(other)):
            return False
        return self.board == other.board

    def printPieces(self):
        board = ""
        for row in reversed(self.data):
            for i in range(-3, 11):
                piece = row.get(i)
                if piece is not None:
                    if piece.color == BLACK:
                        piece_colour = FAN_PIECES[BLACK][piece.piece]
                    else:
                        piece_colour = FAN_PIECES[WHITE][piece.piece]
                    board += piece_colour
                else:
                    board += "."
            board += "\n"
        print(board)



"""Pawn Odds Variant"""

from pychess.Utils.const import PAWNODDSCHESS, VARIANTS_ODDS
from pychess.Utils.Board import Board

PAWNODDSSTART = "rnbqkbnr/ppppp1pp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1"


class PawnOddsBoard(Board):
    """:Description: A standard chess game where one  side starts with one less
    pawn, this is known as giving pawn odds
    """

    variant = PAWNODDSCHESS
    __desc__ = _("One player starts with one less pawn piece")
    name = _("Pawn odds")
    cecp_name = "normal"
    need_initial_board = True
    standard_rules = True
    variant_group = VARIANTS_ODDS

    def __init__(self, setup=False, lboard=None):
        if setup is True:
            Board.__init__(self, setup=PAWNODDSSTART, lboard=lboard)
        else:
            Board.__init__(self, setup=setup, lboard=lboard)



"""Suicide Variation"""

from pychess.Utils.const import (
    SUICIDECHESS,
    VARIANTS_OTHER_NONSTANDARD,
    KING_PROMOTION,
    QUEEN_PROMOTION,
    ROOK_PROMOTION,
    BISHOP_PROMOTION,
    KNIGHT_PROMOTION,
)
from pychess.Utils.Board import Board

SUICIDESTART = "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w - - 0 1"


class SuicideBoard(Board):
    """:Description: This is the FICS version of Losing chess used on FICS as suicide chess.
    You must capture if you can, and the object is to lose all your pieces or to have no moves left.
    But in Suicide, the king is just like any other piece.
    It can move into check and be captured, and you can even promote pawns to kings.
    """

    variant = SUICIDECHESS
    __desc__ = _(
        "FICS suicide: http://www.freechess.org/Help/HelpFiles/suicide_chess.html"
    )
    name = _("Suicide")
    cecp_name = "suicide"
    need_initial_board = True
    standard_rules = False
    variant_group = VARIANTS_OTHER_NONSTANDARD

    PROMOTIONS = (
        KING_PROMOTION,
        QUEEN_PROMOTION,
        ROOK_PROMOTION,
        BISHOP_PROMOTION,
        KNIGHT_PROMOTION,
    )

    def __init__(self, setup=False, lboard=None):
        if setup is True:
            Board.__init__(self, setup=SUICIDESTART, lboard=lboard)
        else:
            Board.__init__(self, setup=setup, lboard=lboard)


def pieceCount(board, color):
    return bin(board.friends[color]).count("1")


if __name__ == "__main__":
    from pychess.Utils.Move import Move
    from pychess.Utils.lutils.lmove import parseAN
    from pychess.Utils.lutils.lmovegen import genCaptures

    FEN = "rnbqk1nr/pppp1pPp/4p3/8/8/8/PPPbPPP1/RNBQKBNR b - - 7 4"
    game = SuicideBoard(SUICIDESTART)

    game = game.move(Move(parseAN(game.board, "h2h4")))
    print(game.board.__repr__())
    for move in genCaptures(game.board):
        print(Move(move))

    game = game.move(Move(parseAN(game.board, "e7e6")))
    print(game.board.__repr__())
    for move in genCaptures(game.board):
        print(Move(move))

    game = game.move(Move(parseAN(game.board, "h4h5")))
    print(game.board.__repr__())
    for move in genCaptures(game.board):
        print(Move(move))

    game = game.move(Move(parseAN(game.board, "f8b4")))
    print(game.board.__repr__())
    for move in genCaptures(game.board):
        print(Move(move))

    game = game.move(Move(parseAN(game.board, "h5h6")))
    print(game.board.__repr__())
    for move in genCaptures(game.board):
        print(Move(move))

    game = game.move(Move(parseAN(game.board, "b4d2")))
    print(game.board.__repr__())
    for move in genCaptures(game.board):
        print(Move(move))

    game = game.move(Move(parseAN(game.board, "h6g7")))
    print(game.board.__repr__())
    for move in genCaptures(game.board):
        print(Move(move))

    game = game.move(Move(parseAN(game.board, "d2e1")))
    print(game.board.__repr__())
    for move in genCaptures(game.board):
        print(Move(move))



import asyncio
import re
from urllib.request import urlopen

from pychess.Utils.lutils.lmovegen import newMove
from pychess.Utils.lutils.lmove import FILE, RANK
from pychess.Utils.const import (
    WHITE,
    DRAW,
    NORMAL_MOVE,
    ENPASSANT,
    EMPTY,
    PAWN,
    BLACKWON,
    WHITEWON,
    QUEEN_PROMOTION,
    ROOK_PROMOTION,
    BISHOP_PROMOTION,
    KNIGHT_PROMOTION,
)

from pychess.Utils.repr import reprColor
from pychess.System.Log import log
from pychess.System import conf

URL = "http://www.k4it.de/egtb/fetch.php?action=egtb&fen="
expression = re.compile(r"(\d+)-(\d+)-?(\d+)?: (Win in \d+|Draw|Lose in \d+)")
PROMOTION_FLAGS = {
    2: QUEEN_PROMOTION,
    3: ROOK_PROMOTION,
    4: BISHOP_PROMOTION,
    5: KNIGHT_PROMOTION,
    8: QUEEN_PROMOTION,
    9: ROOK_PROMOTION,
    10: BISHOP_PROMOTION,
    11: KNIGHT_PROMOTION,
}


class EgtbK4kit:
    def __init__(self):
        self.table = {}

    def supports(self, size):
        return sum(size) < 7

    async def scoreAllMoves(self, board, probeSoft=False):
        global URL, expression, PROMOTION_FLAGS
        fen = board.asFen().split()[0] + " w - - 0 1"
        if (fen, board.color) in self.table:
            return self.table[(fen, board.color)]

        if probeSoft or not conf.get("online_egtb_check"):
            return []

        def get_data(URL, fen):
            # Request the page
            url = (URL + fen).replace(" ", "%20")
            try:
                f = urlopen(url)
            except OSError as e:
                log.warning(
                    "Unable to read endgame tablebase from the Internet: %s" % repr(e)
                )
                data = b""
            data = f.read()
            return data

        loop = asyncio.get_event_loop()
        future = loop.run_in_executor(None, get_data, URL, fen)
        data = await future

        # Parse
        for color, move_data in enumerate(data.split(b"\nNEXTCOLOR\n")):
            try:
                moves = []
                for fcord, tcord, promotion, result in expression.findall(
                    move_data.decode()
                ):
                    fcord = int(fcord)
                    tcord = int(tcord)

                    if promotion:
                        flag = PROMOTION_FLAGS[int(promotion)]
                    elif (
                        RANK(fcord) != RANK(tcord)
                        and FILE(fcord) != FILE(tcord)
                        and board.arBoard[fcord] == PAWN
                        and board.arBoard[tcord] == EMPTY
                    ):
                        flag = ENPASSANT
                    else:
                        flag = NORMAL_MOVE

                    move = newMove(fcord, tcord, flag)

                    if result == "Draw":
                        state = DRAW
                        steps = 0
                    else:
                        s, steps = result.split(" in ")
                        steps = int(steps)

                    if result.startswith("Win"):
                        if color == WHITE:
                            state = WHITEWON
                        else:
                            state = BLACKWON
                    elif result.startswith("Lose"):
                        if color == WHITE:
                            state = BLACKWON
                        else:
                            state = WHITEWON

                    moves.append((move, state, steps))

                if moves:
                    self.table[(fen, color)] = moves
                elif color == board.color and board.opIsChecked():
                    log.warning("Asked endgametable for a won position: %s" % fen)
                elif color == board.color:
                    log.warning(
                        "Couldn't get %s data for position %s.\nData was: %s"
                        % (reprColor[color], fen, repr(data))
                    )
            except (KeyError, ValueError):
                log.warning(
                    "Couldn't parse {} data for position {}.\nData was: {}".format(
                        reprColor[color], fen, repr(data)
                    )
                )
                self.table[(fen, color)] = []  # Don't try again.

        if (fen, board.color) in self.table:
            return self.table[(fen, board.color)]
        return []

    def scoreGame(self, board, omitDepth, probeSoft):
        scores = self.scoreAllMoves(board, probeSoft)
        if scores:
            return scores[0][1], scores[0][2]
        return None, None



from gi.repository import GObject

from pychess.Players.Engine import Engine
from pychess.Utils.const import NORMAL, ANALYZING, INVERSE_ANALYZING

TIME_OUT_SECOND = 60


class ProtocolEngine(Engine):
    __gsignals__ = {
        "readyForOptions": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "readyForMoves": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    # Setting engine options

    def __init__(self, subprocess, color, protover, md5):
        Engine.__init__(self, md5)

        self.engine = subprocess
        self.defname = subprocess.defname
        self.color = color
        self.protover = protover

        self.readyMoves = False
        self.readyOptions = False

        self.connected = True
        self.mode = NORMAL
        self.analyzing_paused = False

    def isAnalyzing(self):
        return self.mode in (ANALYZING, INVERSE_ANALYZING)



from gi.repository import Gtk

from pychess.Utils.Piece import Piece
from pychess.Utils.const import (
    WHITE,
    HAWK,
    ELEPHANT,
    HAWK_GATE,
    ELEPHANT_GATE,
    HAWK_GATE_AT_ROOK,
    ELEPHANT_GATE_AT_ROOK,
)
from pychess.System import uistuff
from pychess.widgets import mainwindow

from .PieceWidget import PieceWidget


class GatingDialog:
    def __init__(self, variant):
        self.widgets = uistuff.GladeWidgets("gating.glade")
        self.widgets["gatingDialog"].set_transient_for(mainwindow())
        self.dialog = self.widgets["gatingDialog"]

        self.color = None

        if self.widgets["hawkDock"].get_child() is None:
            self.widgets["hawkDock"].add(PieceWidget(Piece(WHITE, HAWK), variant))
            self.widgets["hawkDock"].get_child().show()
            self.widgets["elephantDock"].add(
                PieceWidget(Piece(WHITE, ELEPHANT), variant)
            )
            self.widgets["elephantDock"].get_child().show()

            self.widgets["hawkAtRookDock"].add(PieceWidget(Piece(WHITE, HAWK), variant))
            self.widgets["hawkAtRookDock"].get_child().show()
            self.widgets["elephantAtRookDock"].add(
                PieceWidget(Piece(WHITE, ELEPHANT), variant)
            )
            self.widgets["elephantAtRookDock"].get_child().show()

    def setColor(self, color):
        self.widgets["hawkDock"].get_child().getPiece().color = color
        self.widgets["elephantDock"].get_child().getPiece().color = color

        self.widgets["hawkAtRookDock"].get_child().getPiece().color = color
        self.widgets["elephantAtRookDock"].get_child().getPiece().color = color

    def runAndHide(self, color, castling, hawk, elephant):
        self.setColor(color)

        self.widgets["button1"].show()
        self.widgets["button2"].show()
        self.widgets["button3"].show()
        self.widgets["button4"].show()

        if not castling:
            self.widgets["button3"].hide()
            self.widgets["button4"].hide()

        if not hawk:
            self.widgets["button1"].hide()
            self.widgets["button3"].hide()

        if not elephant:
            self.widgets["button2"].hide()
            self.widgets["button4"].hide()

        res = self.dialog.run()
        self.dialog.hide()
        if res != Gtk.ResponseType.DELETE_EVENT:
            return [HAWK_GATE, ELEPHANT_GATE, HAWK_GATE_AT_ROOK, ELEPHANT_GATE_AT_ROOK][
                int(res)
            ]
        return None



import re

from gi.repository import Gtk
from pychess.System import uistuff
from pychess.System.prefix import addDataPrefix

TYPE_PERSONAL, TYPE_CHANNEL, TYPE_GUEST, TYPE_ADMIN, TYPE_COMP, TYPE_BLINDFOLD = range(
    6
)


def get_playername(playername):
    re_m = re.match(r"(\w+)\W*", playername)
    return re_m.groups()[0]


__title__ = _("Talking")

__icon__ = addDataPrefix("glade/panel_chat.svg")

__desc__ = _("List of server channels")


class Sidepanel:
    def load(self, widgets, connection, lounge):
        self.connection = connection

        # deferred imports to not slow down PyChess starting up
        from pychess.widgets.ViewsPanel import ViewsPanel
        from pychess.widgets.InfoPanel import InfoPanel
        from pychess.widgets.ChannelsPanel import ChannelsPanel

        self.viewspanel = ViewsPanel(self.connection)
        self.channelspanel = ChannelsPanel(self.connection)
        self.adj = self.channelspanel.get_vadjustment()
        self.infopanel = InfoPanel(self.connection)

        self.chatbox = Gtk.Paned()
        __widget__ = self.chatbox

        self.chatbox.add1(self.channelspanel)

        notebook = Gtk.Notebook()
        notebook.append_page(self.viewspanel, Gtk.Label(_("Chat")))
        notebook.append_page(self.infopanel, Gtk.Label(_("Info")))
        self.chatbox.add2(notebook)

        self.panels = [self.viewspanel, self.channelspanel, self.infopanel]
        self.viewspanel.connect(
            "channel_content_Changed", self.channelspanel.channel_Highlight, id
        )

        self.channelspanel.connect("conversationAdded", self.onConversationAdded)
        self.channelspanel.connect("conversationRemoved", self.onConversationRemoved)
        self.channelspanel.connect("conversationSelected", self.onConversationSelected)
        self.channelspanel.connect("focus_in_event", self.focus_in, self.adj)

        for panel in self.panels:
            panel.show_all()
            panel.start()
        self.chatbox.show_all()

        uistuff.keep(self.chatbox, "chat_paned_position")

        return __widget__

    def onConversationAdded(self, panel, grp_id, text, grp_type):
        # deferred import to not slow down PyChess starting up
        from pychess.widgets.ChatView import ChatView

        chatView = ChatView()
        plus_channel = "+channel " + str(grp_id)
        self.connection.cm.connection.client.run_command(plus_channel)
        for panel in self.panels:
            panel.addItem(grp_id, text, grp_type, chatView)

    def onConversationRemoved(self, panel, grp_id):
        minus_channel = "-channel " + str(grp_id)
        self.connection.cm.connection.client.run_command(minus_channel)
        for panel in self.panels:
            panel.removeItem(grp_id)

    def onConversationSelected(self, panel, grp_id):
        for panel in self.panels:
            panel.selectItem(grp_id)

    def openChatWithPlayer(self, name):
        cm = self.connection.cm
        self.channelspanel.onPersonMessage(cm, name, "", False, "")

    def focus_in(widget, event, adj):
        alloc = widget.get_allocation()
        if alloc.y < adj.value or alloc.y > adj.value + adj.page_size:
            adj.set_value(min(alloc.y, adj.upper - adj.page_size))



import time
import logging

from .prefix import addUserDataPrefix

newName = time.strftime("%Y-%m-%d_%H-%M-%S") + ".log"
logformat = "%(asctime)s.%(msecs)03d %(task)s %(levelname)s: %(message)s"

# delay=True argument prevents creating empty .log files
file_handler = logging.FileHandler(
    addUserDataPrefix(newName), delay=True, encoding="utf-8"
)


class TaskFormatter(logging.Formatter):
    def __init__(self, fmt=None, datefmt=None):
        logging.Formatter.__init__(self, fmt, datefmt)

    def format(self, record):
        if not hasattr(record, "task"):
            record.task = "unknown"

        record.message = record.getMessage()
        record.asctime = self.formatTime(record, self.datefmt)

        s = self._fmt % record.__dict__

        if record.exc_info:
            # Cache the traceback text to avoid converting it multiple times
            # (it's constant anyway)
            if not record.exc_text:
                record.exc_text = self.formatException(record.exc_info)
        if record.exc_text:
            if s[-1:] != "\n":
                s = s + "\n"
            s = s + record.exc_text

        return s


formatter = TaskFormatter(fmt=logformat, datefmt="%H:%M:%S")
file_handler.setFormatter(formatter)

logger = logging.getLogger()
logger.addHandler(file_handler)


class ExtraAdapter(logging.LoggerAdapter):
    def process(self, msg, kwargs):
        kwargs["extra"] = kwargs.get("extra", {"task": "Default"})
        return msg, kwargs


log = ExtraAdapter(logger, {})


class LoggerWriter:
    def __init__(self, logger, level):
        self.logger = logger
        self.level = level

    def write(self, message):
        if message != "\n":
            self.logger.log(self.level, message)

    def flush(self):
        pass


def setup_glib_logging():
    """Code from https://github.com/GNOME/meld/blob/master/bin/meld"""
    from gi.repository import GLib

    levels = {
        GLib.LogLevelFlags.LEVEL_DEBUG: logging.DEBUG,
        GLib.LogLevelFlags.LEVEL_INFO: logging.INFO,
        GLib.LogLevelFlags.LEVEL_MESSAGE: logging.INFO,
        GLib.LogLevelFlags.LEVEL_WARNING: logging.WARNING,
        GLib.LogLevelFlags.LEVEL_ERROR: logging.ERROR,
        GLib.LogLevelFlags.LEVEL_CRITICAL: logging.CRITICAL,
    }

    # Just to make sphinx happy...
    try:
        level_flag = (
            GLib.LogLevelFlags.LEVEL_WARNING
            | GLib.LogLevelFlags.LEVEL_ERROR
            | GLib.LogLevelFlags.LEVEL_CRITICAL
        )
    except TypeError:
        level_flag = GLib.LogLevelFlags.LEVEL_INFO

    log_domain = "Gtk"
    log = logging.getLogger(log_domain)

    def silence(message):
        if "gtk_container_remove: assertion" in message:
            # Looks like it was some hackish code in GTK+ which is now removed:
            # https://git.gnome.org/browse/gtk+/commit/?id=a3805333361fee37a3b1a974cfa6452a85169f08
            return True
        elif "GdkPixbuf" in message:
            return True
        return False

    # This logging handler is for "old" glib logging using a simple
    # syslog-style API.
    def log_adapter(domain, level, message, user_data):
        if not silence(message):
            log.log(levels.get(level, logging.WARNING), message)

    try:
        GLib.log_set_handler(log_domain, level_flag, log_adapter, None)
    except AttributeError:
        # Only present in glib 2.46+
        pass

    # This logging handler is for new glib logging using a structured
    # API. Unfortunately, it was added in such a way that the old
    # redirection API became a no-op, so we need to hack both of these
    # handlers to get it to work.
    def structured_log_adapter(level, fields, field_count, user_data):
        try:
            message = GLib.log_writer_format_fields(level, fields, True)
        except UnicodeDecodeError:
            for field in fields:
                print(field.key, field.value)
            return GLib.LogWriterOutput.HANDLED

        if not silence(message):
            log.log(levels.get(level, logging.WARNING), message)
        return GLib.LogWriterOutput.HANDLED

    try:
        GLib.log_set_writer_func(structured_log_adapter, None)
    except AttributeError:
        # Only present in glib 2.50+
        pass



import asyncio
import sys
import random
import time
import platform
import subprocess
import getpass
import os
import shutil

from pychess.System.Log import log
from pychess.System.prefix import getEngineDataPrefix
from pychess.ic.icc import B_DTGR_END, B_UNIT_END

if not hasattr(asyncio.StreamReader, "readuntil"):
    from pychess.System.readuntil import readuntil, _wait_for_data

    asyncio.StreamReader.readuntil = readuntil
    asyncio.StreamReader._wait_for_data = _wait_for_data

ENCODE = [ord(i) for i in "Timestamp (FICS) v1.0 - programmed by Henrik Gram."]
ENCODELEN = len(ENCODE)
G_RESPONSE = "\x029"
FILLER = b"1234567890abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
# was: b"".join([telnetlib.IAC, telnetlib.WONT, telnetlib.ECHO])
# but telnetlib was removed in Python 3.13
IAC_WONT_ECHO = b"\xff\xfc\x01"

_DEFAULT_LIMIT = 2**16


class CanceledException(Exception):
    pass


class ICSStreamReader(asyncio.StreamReader):
    def __init__(self, limit, loop, connected_event, name):
        asyncio.StreamReader.__init__(self, limit=limit, loop=loop)
        self.connected_event = connected_event
        self.name = name

    async def read_until(self, *untils):
        while True:
            for i, until in enumerate(untils):
                start = self._buffer.find(bytearray(until, "ascii"))
                if start >= 0:
                    self._buffer = self._buffer[start:]
                    return i
                else:
                    not_find = f"read_until:{until} , got:'{self._buffer}'"
                    log.debug(not_find, extra={"task": (self.name, "raw")})

            await self._wait_for_data("read_until")


class ICSStreamReaderProtocol(asyncio.StreamReaderProtocol):
    def __init__(self, stream_reader, client_connected_cb, loop, name, timeseal):
        asyncio.StreamReaderProtocol.__init__(
            self, stream_reader, client_connected_cb=client_connected_cb, loop=loop
        )
        self.name = name
        self.timeseal = timeseal
        self.connected = False
        self.stateinfo = None

    def data_received(self, data):
        cooked = self.cook_some(data)
        self._stream_reader.feed_data(cooked)

    def cook_some(self, data):
        if not self.connected:
            log.debug(data, extra={"task": (self.name, "raw")})
            self.connected = True
            if b"FatICS" in data:
                self.FatICS = True
            elif b"puertorico.com" in data:
                self.USCN = True
                data = data.replace(IAC_WONT_ECHO, b"")
            elif b"chessclub.com" in data:
                self.ICC = True
                data = data.replace(IAC_WONT_ECHO, b"")
            elif b"Starting FICS session" in data:
                data = data.replace(IAC_WONT_ECHO, b"")
        else:
            if self.timeseal:
                data, g_count, self.stateinfo = self.decode(data, self.stateinfo)
            data = data.replace(b"\r", b"").replace(b"\x07", b"")
            # enable this only for temporary debugging
            log.debug(data, extra={"task": (self.name, "raw")})

            if self.timeseal:
                for i in range(g_count):
                    self._stream_reader.stream_writer.write(
                        self.encode(bytearray(G_RESPONSE, "utf-8")) + b"\n"
                    )
        return data

    def encode(self, inbuf, timestamp=None):
        assert inbuf == b"" or inbuf[-1] != b"\n"

        if not timestamp:
            timestamp = int(time.time() * 1000 % 1e7)
        enc = inbuf + bytearray("\x18%d\x19" % timestamp, "ascii")
        padding = 12 - len(enc) % 12
        filler = random.sample(FILLER, padding)
        enc += bytearray(filler)

        buf = enc

        for i in range(0, len(buf), 12):
            buf[i + 11], buf[i] = buf[i], buf[i + 11]
            buf[i + 9], buf[i + 2] = buf[i + 2], buf[i + 9]
            buf[i + 7], buf[i + 4] = buf[i + 4], buf[i + 7]

        encode_offset = random.randrange(ENCODELEN)

        for i in range(len(buf)):
            buf[i] |= 0x80
            j = (i + encode_offset) % ENCODELEN
            buf[i] = (buf[i] ^ ENCODE[j]) - 32

        buf += bytearray([0x80 | encode_offset])
        return buf

    def decode(self, buf, stateinfo=None):
        expected_table = b"[G]\n\r"
        # TODO: add support to FatICS's new zipseal protocol when it finalizes
        # expected_table = "[G]\n\r" if not self.FatICS else "[G]\r\n"
        final_state = len(expected_table)
        g_count = 0
        result = []

        if stateinfo:
            state, lookahead = stateinfo
        else:
            state, lookahead = 0, []

        lenb = len(buf)
        idx = 0
        while idx < lenb:
            buffer_item = buf[idx]
            expected = expected_table[state]
            if buffer_item == expected:
                state += 1
                if state == final_state:
                    g_count += 1
                    lookahead = []
                    state = 0
                else:
                    lookahead.append(buffer_item)
                idx += 1
            elif state == 0:
                result.append(buffer_item)
                idx += 1
            else:
                result.extend(lookahead)
                lookahead = []
                state = 0

        return bytearray(result), g_count, (state, lookahead)


# You can get ICC timestamp from
# https://www.chessclub.com/user/resources/icc/timestamp/
if sys.platform == "win32":
    timestamp = "timestamp_win32.exe"
else:
    timestamp = "timestamp_linux_2.6.8"

altpath = getEngineDataPrefix()
timestamp_path = shutil.which(timestamp, os.X_OK, path=altpath)


class ICSTelnet:
    sensitive = False

    def __init__(self, timeseal):
        self.name = ""
        self.connected = False
        self.canceled = False
        self.FatICS = False
        self.USCN = False
        self.ICC = False
        self.timeseal = timeseal
        self.ICC_buffer = ""

    async def start(self, host, port, connected_event):
        if self.canceled:
            raise CanceledException()
        self.port = port
        self.host = host
        self.connected_event = connected_event

        self.name = host

        if host == "chessclub.com":
            self.ICC = True

            if self.timeseal and timestamp_path is not None:
                self.host = "127.0.0.1"
                self.port = 5500
                try:
                    if sys.platform == "win32":
                        # To prevent engines opening console window
                        startupinfo = subprocess.STARTUPINFO()
                        startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
                    else:
                        startupinfo = None
                    create = asyncio.create_subprocess_exec(
                        *["%s" % timestamp_path, "-p", "%s" % self.port],
                        startupinfo=startupinfo,
                    )
                    self.timestamp_proc = await create
                    log.info("%s started OK" % timestamp_path)
                except OSError as err:
                    log.info(
                        "Can't start {} OSError: {} {}".format(
                            timestamp_path, err.errno, err.strerror
                        )
                    )
                    self.port = port
                    self.host = host
            else:
                log.info("%s not found" % timestamp_path)

            self.timeseal = False

        def cb(reader, writer):
            reader.stream_writer = writer
            reader.connected_event.set()

        loop = asyncio.get_event_loop()
        self.reader = ICSStreamReader(
            _DEFAULT_LIMIT, loop, self.connected_event, self.name
        )
        self.protocol = ICSStreamReaderProtocol(
            self.reader, cb, loop, self.name, self.timeseal
        )
        coro = loop.create_connection(lambda: self.protocol, self.host, self.port)
        self.transport, _protocol = await coro
        # writer = asyncio.StreamWriter(transport, protocol, reader, loop)

        if self.timeseal:
            self.write(self.get_init_string())

    def cancel(self):
        self.canceled = True
        self.close()

    def close(self):
        if self.protocol.connected:
            self.protocol.connected = False
            self.transport.close()

    def get_init_string(self):
        """timeseal header: TIMESTAMP|bruce|Linux gruber 2.6.15-gentoo-r1 #9
        PREEMPT Thu Feb 9 20:09:47 GMT 2006 i686 Intel(R) Celeron(R) CPU
        2.00GHz GenuineIntel GNU/Linux| 93049"""
        user = getpass.getuser()
        uname = " ".join(list(platform.uname()))
        return "TIMESTAMP|" + user + "|" + uname + "|"

    def write(self, string):
        logstr = "*" * len(string) if self.sensitive else string
        self.sensitive = False
        log.info(logstr, extra={"task": (self.name, "raw")})

        if self.timeseal:
            self.transport.write(
                self.protocol.encode(bytearray(string, "utf-8")) + b"\n"
            )
        else:
            self.transport.write(string.encode() + b"\n")

    async def readline(self):
        if self.canceled:
            raise CanceledException()

        if self.ICC:
            line = await self.readuntil(b"\n")
            return line.strip()
        else:
            line = await self.reader.readline()
            return line.decode("latin_1").strip()

    async def readuntil(self, until):
        if self.canceled:
            raise CanceledException()

        if self.ICC:
            if len(self.ICC_buffer) == 0:
                self.ICC_buffer = await self.reader.readuntil(until)
            i = self.ICC_buffer.find(until)
            m = sys.maxsize
            if i >= 0:
                m = i
            j = self.ICC_buffer.find(B_DTGR_END)
            if j >= 0:
                m = min(m, j)
            k = self.ICC_buffer.find(B_UNIT_END)
            if k >= 0:
                m = min(m, k)
            if m != sys.maxsize:
                if m == i:
                    stuff = self.ICC_buffer[: m + len(until)]
                    self.ICC_buffer = self.ICC_buffer[m + len(until) :]
                    return stuff.decode("latin_1")
                else:
                    stuff = self.ICC_buffer[: m + 2]
                    self.ICC_buffer = self.ICC_buffer[m + 2 :]
                    return stuff.decode("latin_1")
            else:
                return ""
        else:
            data = await self.reader.readuntil(until)
            return data.decode("latin_1")

    async def read_until(self, *untils):
        if self.canceled:
            raise CanceledException()

        ret = await self.reader.read_until(*untils)
        return ret



import sys

import gi

try:
    gi.require_version("Rsvg", "2.0")
    from gi.repository import Rsvg
except Exception:
    print("Failed to import required gi module version")
    sys.exit(1)

from pychess.Utils.const import (
    BLACK,
    WHITE,
    KING,
    QUEEN,
    BISHOP,
    KNIGHT,
    ROOK,
    PAWN,
    reprSign,
    SITTUYINCHESS,
    HAWK,
    ELEPHANT,
    SCHESS,
)
from pychess.System import conf
from pychess.System.prefix import addDataPrefix


piece_ord = {KING: 0, QUEEN: 1, ROOK: 2, BISHOP: 3, KNIGHT: 4, PAWN: 5}
pnames = ("Pawn", "Knight", "Bishop", "Rook", "Queen", "King")

size = 800.0


def drawPiece(
    piece,
    context,
    x,
    y,
    psize,
    drawAsWhite=False,
    drawAsPawn=False,
    asean=False,
    variant=None,
):
    """Rendering pieces using .svg chess figurines"""

    color = WHITE if drawAsWhite else piece.color
    sign = PAWN if drawAsPawn else piece.sign
    if variant is not None and variant == SITTUYINCHESS:
        image = sittuyin_svg_pieces[color][sign]
        w, h = image.props.width, image.props.height
        offset_x = 0
        offset_y = 0
    elif asean:
        image = makruk_svg_pieces[color][sign]
        w, h = image.props.width, image.props.height
        offset_x = 0
        offset_y = 0
    elif all_in_one:
        image = svg_pieces
        w, h = image.props.width / 6, image.props.height / 2
        offset_x = piece_ord[sign] * psize
        offset_y = 0 if color == BLACK else psize
    else:
        if variant is not None and variant == SCHESS:
            image = schess_svg_pieces[color][sign]
        else:
            image = svg_pieces[color][sign]
        w, h = image.props.width, image.props.height
        offset_x = 0
        offset_y = 0

    context.save()

    context.rectangle(x, y, psize, psize)
    context.clip()
    context.translate(x - offset_x, y - offset_y)
    context.scale(1.0 * psize / w, 1.0 * psize / h)

    context.push_group()

    if asean:
        image.render_cairo(context)
    elif all_in_one:
        pieceid = "#{}{}".format("White" if color == 0 else "Black", pnames[sign - 1])
        image.render_cairo_sub(context, id=pieceid)
    else:
        image.render_cairo(context)

    context.pop_group_to_source()
    context.paint_with_alpha(piece.opacity)
    context.restore()


surfaceCache = {}

pieces = (PAWN, KNIGHT, BISHOP, ROOK, QUEEN, KING, HAWK, ELEPHANT)


def get_svg_pieces(svgdir):
    """Load figurines from .svg files"""

    if all_in_one:
        rsvg_handles = Rsvg.Handle.new_from_file(
            addDataPrefix(f"pieces/{svgdir}/{svgdir}.svg")
        )
    else:
        rsvg_handles = [[None] * 9, [None] * 9]
        for c, color in ((WHITE, "white"), (BLACK, "black")):
            for p in pieces:
                if p in (HAWK, ELEPHANT) and svgdir != "merida":
                    continue
                rsvg_handles[c][p] = Rsvg.Handle.new_from_file(
                    addDataPrefix(
                        "pieces/{}/{}{}.svg".format(
                            svgdir, color[0], reprSign[p].lower()
                        )
                    )
                )
    return rsvg_handles


all_in_one = None
svg_pieces = None
makruk_svg_pieces = get_svg_pieces("makruk")
sittuyin_svg_pieces = get_svg_pieces("sittuyin")
schess_svg_pieces = get_svg_pieces("merida")
piece2char = None


def set_piece_theme(piece_set):
    global all_in_one
    global svg_pieces
    global piece2char

    piece_set = piece_set.lower()
    if piece_set in (
        "celtic",
        "eyes",
        "fantasy",
        "fantasy_alt",
        "freak",
        "prmi",
        "skulls",
        "spatial",
    ):
        all_in_one = True
    else:
        all_in_one = False

    try:
        svg_pieces = get_svg_pieces(piece_set)
    except Exception:
        print("Can't create piece set %s" % piece_set)
        all_in_one = False
        svg_pieces = get_svg_pieces("merida")


set_piece_theme(conf.get("pieceTheme"))



import collections

from .ChessFile import ChessFile
from pychess.Utils.Cord import Cord
from pychess.Utils.GameModel import GameModel
from pychess.Utils.lutils.LBoard import LBoard
from pychess.Utils.const import (
    DRAW,
    WHITEWON,
    BLACKWON,
    WAITING_TO_START,
    WHITE,
    BLACK,
    KING,
    QUEEN,
    ROOK,
    BISHOP,
    KNIGHT,
    PAWN,
)

__label__ = _("Chess Compositions from yacpdb.org")
__ending__ = "olv"
__append__ = True


# Note S for the knights as N is reserverd for Nightriders
chr2piece = {
    "K": KING,
    "Q": QUEEN,
    "R": ROOK,
    "B": BISHOP,
    "S": KNIGHT,
    "P": PAWN,
}


def save(handle, model, position=None):
    pass


def load(handle):
    return OLVFile(handle)


class OLVFile(ChessFile):
    def __init__(self, handle):
        ChessFile.__init__(self, handle)

        self.games = self.read_games(handle)
        self.count = len(self.games)

    def read_games(self, handle):
        """We don't return games if stipulation is not 'mate in #'"""
        games = []
        rec = None
        rec_id = 1
        contains_fairy_pieces = False

        # authors is a more line list (each line starts with "-")
        in_authors = False

        # piece list are usually in one line list inside []
        # but sometimes given in more line lists
        in_white = False
        in_black = False

        for line in handle:
            line = line.rstrip()

            if in_authors and ":" in line:
                in_authors = False

            elif in_white and ":" in line:
                in_white = False

            elif in_black and ":" in line:
                in_black = False
                rec["FEN"] = self.lboard.asFen()

            # New record start
            if line == "---":
                if (
                    rec is not None
                    and rec["Black"].startswith("Mate in ")
                    and not contains_fairy_pieces
                ):
                    games.append(rec)
                    rec_id += 1

                contains_fairy_pieces = False

                self.lboard = LBoard()
                self.lboard.applyFen("8/8/8/8/8/8/8/8 w - - 0 1")

                rec = collections.defaultdict(str)
                rec["Id"] = rec_id
                rec["Offset"] = 0

            elif line.startswith("authors:"):
                in_authors = True

            elif line.startswith("source:"):
                rec["Event"] = line[8:]

            elif line.startswith("source-id:"):
                rec["Event"] = "{} ({})".format(rec["Event"], line[12:])

            elif line.startswith("date:"):
                parts = line[6:].split("-")
                parts_len = len(parts)
                if parts_len >= 3:
                    rec["Day"] = parts[2]
                if parts_len >= 2:
                    rec["Month"] = parts[1]
                if parts_len >= 1:
                    rec["Year"] = parts[0]

            elif line.startswith("distinction:"):
                rec["Site"] = line[12:]

            elif line.startswith("algebraic:"):
                pass

            elif line.startswith("  white:"):
                parts = line.split("[")
                if len(parts) > 1:
                    pieces = parts[1][:-1]
                    for piece in pieces.split(", "):
                        if piece.startswith("Royal") or piece[0] not in chr2piece:
                            contains_fairy_pieces = True
                        else:
                            cord = Cord(piece[1:3]).cord
                            piece = chr2piece[piece[0]]
                            self.lboard._addPiece(cord, piece, WHITE)
                else:
                    in_white = True

            elif line.startswith("  black:"):
                parts = line.split("[")
                if len(parts) > 1:
                    pieces = parts[1][:-1]
                    for piece in pieces.split(", "):
                        if piece.startswith("Royal") or piece[0] not in chr2piece:
                            contains_fairy_pieces = True
                        else:
                            cord = Cord(piece[1:3]).cord
                            piece = chr2piece[piece[0]]
                            self.lboard._addPiece(cord, piece, BLACK)

                    rec["FEN"] = self.lboard.asFen()
                else:
                    in_black = True

            elif line.startswith("stipulation:"):
                if line.endswith("Black to move"):
                    line = line[:-14]
                    rec["FEN"] = rec["FEN"].replace("w", "b")

                line = line.split(": ")[1]
                if "+" in line:
                    rec["Result"] = WHITEWON
                    rec["Black"] = "Win"
                elif "-" in line:
                    rec["Result"] = BLACKWON
                    rec["Black"] = "Win"
                elif "=" in line:
                    rec["Result"] = DRAW
                    rec["Black"] = "Draw"
                elif line.startswith('"#'):
                    rec["Result"] = WHITEWON
                    rec["Black"] = "Mate in %s" % line[2:-1]
                    rec["Termination"] = "mate in %s" % line[2:-1]

            elif line.startswith("solution:"):
                # TODO: solutions can be in several (sometimes rather unusual) form
                pass

            else:
                if in_authors:
                    author = line[line.find("-") + 1 :].lstrip()
                    if rec["White"]:
                        rec["White"] = "{} - {}".format(rec["White"], author)
                    else:
                        rec["White"] = author

                elif in_white:
                    piece = line[line.find("-") + 1 :].lstrip()
                    cord = Cord(piece[1:3]).cord
                    piece = chr2piece[piece[0]]
                    self.lboard._addPiece(cord, piece, WHITE)

                elif in_black:
                    piece = line[line.find("-") + 1 :].lstrip()
                    cord = Cord(piece[1:3]).cord
                    piece = chr2piece[piece[0]]
                    self.lboard._addPiece(cord, piece, BLACK)

        # Append the latest record
        if (
            rec is not None
            and rec["Black"].startswith("Mate in ")
            and not contains_fairy_pieces
        ):
            games.append(rec)

        return games

    def loadToModel(self, rec, position, model=None):
        if not model:
            model = GameModel()

        model.tags["Event"] = rec["Event"]
        model.tags["Site"] = rec["Site"]
        model.tags["Date"] = self.get_date(rec)
        model.tags["Round"] = ""
        model.tags["White"] = "?"
        model.tags["Black"] = "?"
        model.tags["Termination"] = rec["Termination"]
        fen = rec["FEN"]
        model.tags["FEN"] = fen

        model.boards = [model.variant(setup=fen)]
        model.variations = [model.boards]
        model.status = WAITING_TO_START

        return model

    def get_date(self, rec):
        year = rec["Year"]
        month = rec["Month"]
        day = rec["Day"]
        if year and month and day:
            tag_date = "%s.%02d.%02d" % (year, int(month), int(day))
        elif year and month:
            tag_date = "%s.%02d" % (year, int(month))
        elif year:
            tag_date = "%s" % year
        else:
            tag_date = ""
        return tag_date



import os
from struct import Struct
from collections import namedtuple

from pychess.System import conf
from pychess.Utils.lutils.lmove import parsePolyglot
from pychess.System.Log import log

path = conf.get("opening_file_entry")

if os.path.isfile(path):
    bookfile = True
else:
    bookfile = False
    log.warning("Could not find %s" % path)

# The book probing code is based on that of PolyGlot by Fabien Letouzey.
# PolyGlot is available under the GNU GPL from http://wbec-ridderkerk.nl

BookEntry = namedtuple("BookEntry", "key move weight learn")
# 'key' c_uint64      the position's hash
# 'move' c_uint16     the candidate move
# 'weight' c_uint16   proportional to prob. we should play it
# The following terms are not always available:
# 'learn' c_uint32    we use this NOT the polyglot way but as in
#                     https://github.com/mcostalba/chess_db

entrystruct = Struct(">QHHI")
entrysize = entrystruct.size


def getOpenings(board):
    """Return a tuple (move, weight, learn) for each opening move
    in the given position. The weight is proportional to the probability
    that a move should be played. By convention, games is the number of
    times a move has been tried, and score the number of points it has
    scored (with 2 per victory and 1 per draw). However, opening books
    aren't required to keep this information."""

    openings = []
    if not bookfile:
        return openings

    with open(path, "rb") as bookFile:
        key = board.hash
        # Find the first entry whose key is >= the position's hash
        bookFile.seek(0, os.SEEK_END)
        low, high = 0, bookFile.tell() // 16 - 1
        if high < 0:
            return openings
        while low < high:
            mid = (low + high) // 2
            bookFile.seek(mid * 16)
            entry = bookFile.read(entrysize)
            if len(entry) != entrysize:
                return openings
            entry = BookEntry._make(entrystruct.unpack(entry))
            if entry.key < key:
                low = mid + 1
            else:
                high = mid

        bookFile.seek(low * 16)
        while True:
            entry = bookFile.read(entrysize)
            if len(entry) != entrysize:
                return openings
            entry = BookEntry._make(entrystruct.unpack(entry))
            if entry.key != key:
                break
            move = parsePolyglot(board, entry.move)
            openings.append((move, entry.weight, entry.learn))
    return openings



# Authors: Jonas Thiem

import re

from gi.repository import Gtk, GObject, Pango

from pychess.System.Log import log
from pychess.System.prefix import addDataPrefix
from pychess.Utils.const import ARTIFICIAL

__title__ = _("Engines")

__icon__ = addDataPrefix("glade/panel_engineoutput.svg")
white = addDataPrefix("glade/panel_engineoutput.svg")

__desc__ = _(
    "The engine output panel shows the thinking output of chess engines (computer players) during a game"
)


class Sidepanel:
    def load(self, gmwidg):
        # Specify whether the panel should have a horizontal layout:
        horizontal = True

        if horizontal:
            self.box = Gtk.HBox()
        else:
            self.box = Gtk.VBox()
        __widget__ = self.box

        # Use two engine output widgets for each player color:
        self.output_white = EngineOutput(True)
        self.output_black = EngineOutput(False)
        if horizontal:
            self.output_separator = Gtk.VSeparator()
        else:
            self.output_separator = Gtk.HSeparator()

        self.output_noengines = Gtk.TextView()
        self.output_noengines.get_buffer().set_text(
            _("No chess engines (computer players) are participating in this game.")
        )
        self.output_noengines.set_editable(False)
        self.output_noengines.set_wrap_mode(Gtk.WrapMode.WORD_CHAR)

        __widget__.pack_start(self.output_noengines, True, True, 0)
        __widget__.show_all()

        self.boardview = gmwidg.board.view

        self.model_cids = [
            self.boardview.model.connect_after("game_changed", self.game_changed),
            self.boardview.model.connect_after("players_changed", self.players_changed),
            self.boardview.model.connect_after("game_started", self.game_started),
            self.boardview.model.connect_after(
                "game_terminated", self.on_game_terminated
            ),
        ]
        return __widget__

    def on_game_terminated(self, model):
        for cid in self.model_cids:
            self.boardview.model.disconnect(cid)

    def updateVisibleOutputs(self, model):
        # Check which players participate and update which views are visible
        # gotplayers = False
        if model is None:
            return
        # gotEnginePlayers = False
        gotWhiteEngine = False
        gotBlackEngine = False
        if len(model.players) > 0:
            if model.players[0].__type__ == ARTIFICIAL:
                gotWhiteEngine = True
                self.output_white.attachEngine(model.players[0].engine)
            if model.players[1].__type__ == ARTIFICIAL:
                gotBlackEngine = True
                self.output_black.attachEngine(model.players[1].engine)

        # First, detach from old engines:
        if not gotBlackEngine:
            self.output_black.detachEngine()
        if not gotWhiteEngine:
            self.output_white.detachEngine()

        if gotBlackEngine or gotWhiteEngine:
            # Remove "no engines" label:
            if self.output_noengines in self.box.get_children():
                self.box.remove(self.output_noengines)

            # Add white engine info if white engine is participating:
            if gotWhiteEngine:
                if self.output_white not in self.box.get_children():
                    # Remove black output and separator first
                    # to ensure proper ordering:
                    if self.output_black in self.box.get_children():
                        self.box.remove(self.output_black)
                        self.box.remove(self.output_separator)
                    self.box.pack_start(self.output_white, True, True, 0)
                    self.output_white.clear()
                    self.output_white.show_all()
                self.output_white.setTitle(model.players[0].name)
            else:
                if self.output_white in self.box.get_children():
                    self.box.remove(self.output_white)
                    self.box.remove(self.output_separator)

            # Add white engine info if black engine is participating:
            if gotBlackEngine:
                if self.output_black not in self.box.get_children():
                    if gotWhiteEngine:
                        self.box.pack_start(self.output_separator, False, True, 0)
                        self.output_separator.show()
                    self.box.pack_start(self.output_black, True, True, 0)
                    self.output_black.clear()
                    self.output_black.show_all()
                self.output_black.setTitle(model.players[1].name)
            else:
                if self.output_black in self.box.get_children():
                    self.box.remove(self.output_black)
                    self.box.remove(self.output_separator)
        else:
            # Show "no engines" label
            if self.output_white in self.box.get_children():
                self.box.remove(self.output_white)
            if self.output_black in self.box.get_children():
                self.box.remove(self.output_black)
            if self.output_noengines not in self.box.get_children():
                self.box.pack_start(self.output_noengines, True, True, 0)
        return

    def players_changed(self, model):
        log.debug("engineOutputPanel.players_changed: starting")
        self.updateVisibleOutputs(model)
        log.debug("engineOutputPanel.players_changed: returning")
        return

    def game_started(self, model):
        self.updateVisibleOutputs(model)
        return

    def game_changed(self, model, ply):
        self.updateVisibleOutputs(model)
        return


class EngineOutput(Gtk.VBox):
    def __init__(self, white=True):
        GObject.GObject.__init__(self)

        self.attached_engine = None  # engine attached to which we listen
        self.white = white
        self.clear_on_output = False  # next thinking line belongs to new move

        # Title bar:
        self.title_label = Gtk.Label()
        self.title_color = Gtk.Image()

        self.title_hbox = Gtk.HBox()
        # self.title_hbox.pack_start(self.title_color, False)
        self.title_hbox.pack_start(self.title_color, False, False, 0)
        # self.title_hbox.pack_start(self.title_label, True, True)
        self.title_hbox.pack_start(self.title_label, True, True, 0)

        # Set black or white player icon in front:
        if white is True:
            self.title_color.set_from_file(addDataPrefix("glade/white.png"))
        else:
            self.title_color.set_from_file(addDataPrefix("glade/black.png"))

        # output scrolled window container:
        self.output_container = Gtk.ScrolledWindow()
        self.output_container.set_policy(
            Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC
        )

        # scroll down on new output: -- not reliable with multilines added
        # uistuff.keepDown(self.output_container)

        # scroll down on new output: -- brute force variant
        def changed(vadjust):
            vadjust.set_value(vadjust.get_upper() - vadjust.get_page_size())

        self.output_container.get_vadjustment().connect("changed", changed)

        # Text field for output:
        self.output = Gtk.TextView()
        self.output_container.add(self.output)
        self.output.set_editable(False)
        self.output.set_wrap_mode(Gtk.WrapMode.WORD_CHAR)
        self.tag_bold = self.output.get_buffer().create_tag(
            "bold", weight=Pango.Weight.BOLD
        )
        self.tag_color = self.output.get_buffer().create_tag(
            "color", foreground="#0033ff"
        )

        # Add all sub widgets to ourselves:
        # self.pack_start(self.title_hbox, False)
        # self.pack_start(self.output_container, True)
        self.pack_start(self.title_hbox, False, False, 0)
        self.pack_start(self.output_container, True, True, 0)

        # Precompile regexes we want to use:
        self.re_thinking_line_cecp = re.compile(r"^[0-9]+\.? +\-?[0-9]+ +")
        self.re_thinking_line_uci = re.compile(
            r"^info (.*) pv [a-hA-H][0-9][a-hA-H][0-9](.*)$"
        )
        self.re_move_line_cecp_alg = re.compile(
            r"^(move +)?[a-hA-H][0-9][a-hA-H][0-9]$"
        )
        self.re_move_line_cecp_san = re.compile(
            r"^(move +)?([QKNB]?[a-hA-H]?[xX@]?[a-hA-H][0-9]\+?#?|[oO]-[oO]-[oO]|[oO]-[oO])$"
        )
        self.re_move_line_uci = re.compile(
            r"^bestmove +[a-hA-H][0-9][a-hA-H][0-9]( .*)?$"
        )
        self.re_extract_cecp_all = re.compile(
            r"^([0-9]+)\.? +(\-?[0-9]+) +([0-9]+).?([0-9]*) ([^ ].*)$"
        )
        self.re_extract_uci_depth = re.compile(r"depth +([0-9]+) +")
        self.re_extract_uci_score = re.compile(r"score cp +(-?[0-9]+) +")
        self.re_extract_uci_score_mate_other = re.compile(r"score +mate +([0-9]+) +")
        self.re_extract_uci_score_mate_us = re.compile(r"score +mate +\-([0-9]+) +")
        self.re_extract_uci_score_lowerbound = re.compile(r"score +lowerbound +")
        self.re_extract_uci_score_upperbound = re.compile(r"score +upperbound +")
        self.re_extract_uci_pv = re.compile(r"pv +([a-hA-HoO].*[^ ]) *$")
        self.re_extract_uci_nps = re.compile(r"nps +([0-9]+) +")

    def _del(self):
        self.detachEngine()

    def appendNewline(self):
        # Start a new line if text output isn't empty:
        if self.output.get_buffer().get_char_count() > 0:
            # We have old content, append newline
            self.output.get_buffer().insert(
                self.output.get_buffer().get_end_iter(), "\n"
            )

    def append(self, line, tag=None):
        # Append a specific string with the given formatting:
        # oldenditer = self.output.get_buffer().get_end_iter()
        self.output.get_buffer().insert(self.output.get_buffer().get_end_iter(), line)
        if tag is not None:
            enditer = self.output.get_buffer().get_end_iter()
            startiter = enditer.copy()
            startiter.backward_chars(len(line))
            self.output.get_buffer().apply_tag(tag, startiter, enditer)

    def appendThinking(self, depth, score, pv, nps):
        # Append a formatted thinking line:
        self.appendNewline()
        self.append(depth.__str__() + ". ", self.tag_color)
        self.append("Score: ", self.tag_bold)
        self.append(score.__str__() + " ")
        self.append("PV: ", self.tag_bold)
        self.append(pv.__str__() + " ")
        if nps != "":
            self.append("NPS: ", self.tag_bold)
            self.append(nps.__str__())

    def parseInfoLine(self, line):
        # Parse an identified info line and add it to our output:
        if self.clear_on_output is True:
            self.clear_on_output = False
            self.clear()

        # Clean up line first:
        while line.find("  ") != -1:
            line = line.replace("  ", " ")

        depth = "?"
        score = "?"
        pv = "?"
        nps = ""
        infoFound = False

        # do more sophisticated parsing here:
        if line.startswith("info "):
            # UCI info line
            # always end with a space to faciliate searching:
            line = line + " "

            # parse depth:
            result = self.re_extract_uci_depth.search(line)
            if result:
                depth = result.group(1)

            # parse score:
            result = self.re_extract_uci_score.search(line)
            if result:
                score = result.group(1)
            else:
                result = self.re_extract_uci_score_mate_other.search(line)
                if result:
                    score = "winning in " + result.group(1) + " moves"
                else:
                    result = self.re_extract_uci_score_mate_us.search(line)
                    if result:
                        score = "losing in " + result.group(1) + " moves"
                    else:
                        if self.re_extract_uci_score_lowerbound.search(line):
                            score = "lowerbound"
                        elif self.re_extract_uci_score_upperbound.search(line):
                            score = "upperbound"
            # parse pv:
            result = self.re_extract_uci_pv.search(line)
            if result:
                infoFound = True
                pv = result.group(1)

            # parse nps:
            result = self.re_extract_uci_nps.search(line)
            if result:
                infoFound = True
                nps = result.group(1)
        else:
            # CECP/Winboard/GNUChess info line
            # parse all information in one go:
            result = self.re_extract_cecp_all.match(line)
            if not result:
                return
            infoFound = True
            depth = result.group(1)
            score = result.group(2)
            time = result.group(3)
            nodes = result.group(4)
            pv = result.group(5)

            nps = str(int(int(nodes) / (int(time) / 100))) if int(time) > 0 else ""
            # Clean pv of unwanted chars:
        pv = re.sub("[^a-z^A-Z^0-9^ ^x^@^?]", "", pv)

        # If we found useful information, show it:
        if infoFound:
            self.appendThinking(depth, score, pv, nps)

    def parseLine(self, engine, line):
        # Clean up the line a bit:
        line = line.strip(" \r\t\n")
        line = line.replace("\t", " ")

        # PARSING THINKING OUTPUT (roughly, simply identifies the lines):

        # GNU Chess/CECP/Winboard engine thinking output lines:
        if self.re_thinking_line_cecp.match(line):
            self.parseInfoLine(line)

        # UCI engine thinking output lines:
        if self.re_thinking_line_uci.match(line):
            if line.find("depth") != -1 and line.find("score") != -1:
                self.parseInfoLine(line)

        # PARSE MOVE LINES (roughly, we merely identify them):

        # We want to clear on the next output info line
        # when a move arrived, so that for every move
        # we freshly fill our thinking output:

        # CECP/Winboard move line, long algebraeic notation:
        if self.re_move_line_cecp_alg.match(line):
            self.clear_on_output = True

        # CECP/Winboard move line, SAN notation:
        if self.re_move_line_cecp_san.match(line):
            self.clear_on_output = True

        # UCI move line:
        if self.re_move_line_uci.match(line):
            self.clear_on_output = True
        return

    def clear(self):
        self.output.get_buffer().set_text("")
        return

    def setTitle(self, title):
        self.title_label.set_text(title)
        return

    def attachEngine(self, engine):
        # Attach an engine for line listening
        if self.attached_engine is not None:
            if self.attached_engine == engine:
                # We are already attached to this engine
                return
            # Detach from previous engine
            self.attached_engine.disconnect(self.attached_handler_id)
        # Attach to new engine:
        log.debug(
            "Attaching " + self.__str__() + " to engine " + engine.__str__(),
            extra={"task": engine.defname},
        )
        self.attached_engine = engine
        self.attached_handler_id = engine.connect("line", self.parseLine)
        return

    def detachEngine(self):
        # Detach from attached engine
        if self.attached_engine is not None:
            log.debug(
                "Detaching "
                + self.__str__()
                + " from engine "
                + self.attached_engine.__str__(),
                extra={"task": self.attached_engine.defname},
            )
            self.attached_engine.disconnect(self.attached_handler_id)
            self.attached_engine = None

    def __repr__(self):
        color = "black"
        if self.white:
            color = "white"
        return "Engine Output " + color + " #" + id(self).__str__()


#    def __str__(self):
#        return repr(self) + " (engine: " + self.attached_engine.__str__() + ")"



import shutil
import collections
import os
from io import StringIO
from os.path import getmtime
import re
import textwrap
import sys

import pexpect

from sqlalchemy import String

from pychess.external.scoutfish import Scoutfish
from pychess.external.chess_db import Parser

from pychess.Utils.const import (
    WHITE,
    BLACK,
    reprResult,
    FEN_START,
    FEN_EMPTY,
    WON_RESIGN,
    DRAW,
    BLACKWON,
    WHITEWON,
    NORMALCHESS,
    DRAW_AGREE,
    FIRST_PAGE,
    PREV_PAGE,
    NEXT_PAGE,
    ABORTED_REASONS,
    ADJOURNED_REASONS,
    ADJUDICATION_REASONS,
    WON_ADJUDICATION,
    DEATH_REASONS,
    CALLFLAG_REASONS,
    RUNNING,
    TOOL_NONE,
    TOOL_CHESSDB,
    TOOL_SCOUTFISH,
)
from pychess.System import conf
from pychess.System.Log import log
from pychess.System.protoopen import PGN_ENCODING
from pychess.System.prefix import getEngineDataPrefix
from pychess.Utils.lutils.LBoard import LBoard
from pychess.Utils.GameModel import GameModel
from pychess.Utils.lutils.lmove import toSAN, parseAny, ParsingError
from pychess.Utils.Move import Move
from pychess.Utils.elo import get_elo_rating_change_pgn
from pychess.Utils.logic import getStatus
from pychess.Variants import name2variant, NormalBoard, variants
from pychess.Utils import formatTime
from pychess.Savers.ChessFile import ChessFile, LoadingError
from pychess.Savers.database import col2label, TagDatabase, parseDateTag
from pychess.System.cpu import get_cpu
from pychess.Database import model as dbmodel
from pychess.Database.PgnImport import TAG_REGEX, pgn2Const, PgnImport
from pychess.Database.model import (
    game,
    create_indexes,
    drop_indexes,
    metadata,
    ini_schema_version,
)

__label__ = _("Chess Game")
__ending__ = "pgn"
__append__ = True


# token categories
(
    COMMENT_REST,
    COMMENT_BRACE,
    COMMENT_NAG,
    VARIATION_START,
    VARIATION_END,
    RESULT,
    FULL_MOVE,
    MOVE,
    MOVE_COMMENT,
) = range(1, 10)

pattern = re.compile(
    r"""
    (\;.*?[\n\r])        # comment, rest of line style
    |(\{.*?\})           # comment, between {}
    |(\$[0-9]+)          # comment, Numeric Annotation Glyph
    |(\()                # variation start
    |(\))                # variation end
    |(\*|1-0|0-1|1/2)    # result (spec requires 1/2-1/2 for draw, but we want to tolerate simple 1/2 too)
    |(
    ([a-hKkQqRrBNnMmSsFHE][a-hxKQRBNMSFHE1-8+#=\-/]{1,6}
    |[PNBRQMSFK]@[a-h][1-8][+#]?  # drop move
    |o\-?o(?:\-?o)?      # castling notation using letter 'o' with or without '-'
    |O\-?O(?:\-?O)?(?:/[HE][a-h][18]?)?      # castling notation using letter 'O' with or without '-'
    |0\-0(?:\-0)?        # castling notation using zero with required '-'
    |\-\-)               # non standard '--' is used for null move inside variations
    ([\?!]{1,2})*
    )    # move (full, count, move with ?!, ?!)
    """,
    re.VERBOSE | re.DOTALL,
)

move_eval_re = re.compile(
    r"\[%eval\s+([+\-])?(?:#)?(\d+)(?:[,\.](\d{1,2}))?(?:/(\d{1,2}))?\]"
)
move_time_re = re.compile(r"\[%emt\s+(\d:)?(\d{1,2}:)?(\d{1,4})(?:\.(\d{1,3}))?\]")

# Chessbase style circles/arrows {[%csl Ra3][%cal Gc2c3,Rc3d4]}
comment_circles_re = re.compile(r"\[%csl\s+((?:[RGBY]\w{2},?)+)\]")
comment_arrows_re = re.compile(r"\[%cal\s+((?:[RGBY]\w{4},?)+)\]")

# Mandatory tags (except "Result")
mandatory_tags = ("Event", "Site", "Date", "Round", "White", "Black")


def msToClockTimeTag(ms):
    """
    Converts milliseconds to a chess clock time string in 'WhiteClock'/
    'BlackClock' PGN header format
    """
    msec = ms % 1000
    sec = ((ms - msec) % (1000 * 60)) / 1000
    minute = ((ms - sec * 1000 - msec) % (1000 * 60 * 60)) / (1000 * 60)
    hour = ((ms - minute * 1000 * 60 - sec * 1000 - msec) % (1000 * 60 * 60 * 24)) / (
        1000 * 60 * 60
    )
    return "%01d:%02d:%02d.%03d" % (hour, minute, sec, msec)


def parseClockTimeTag(tag):
    """
    Parses 'WhiteClock'/'BlackClock' PGN headers and returns the time the
    player playing that color has left on their clock in milliseconds
    """
    match = re.match(r"(\d{1,2}):(\d\d):(\d\d).(\d{1,3})", tag)
    if match:
        hour, minute, sec, msec = match.groups()
        return (
            int(msec)
            + int(sec) * 1000
            + int(minute) * 60 * 1000
            + int(hour) * 60 * 60 * 1000
        )


def parseTimeControlTag(tag):
    """
    Parses 'TimeControl' PGN header and returns the time and gain the
    players have on game start in seconds
    """
    match = re.match(r"^(\d+)\+?(\-?\d+)?$", tag)
    if match:
        secs, gain = match.groups()
        return int(secs), int(gain) if gain is not None else 0, 0
    else:
        match = re.match(r"^(\d+)\/(\d+)$", tag)
        if match:
            moves, secs = match.groups()
            return int(secs), 0, int(moves)
        else:
            return None


def save(handle, model, position=None, flip=False):
    """Saves the game from GameModel to .pgn"""
    processed_tags = []

    def write_tag(tag, value, roster=False):
        nonlocal processed_tags
        if tag in processed_tags or (not roster and not value):
            return
        try:
            pval = str(value)
            pval = pval.replace("\\", "\\\\")
            pval = pval.replace('"', '\\"')
            print(f'[{tag} "{pval}"]', file=handle)
        except UnicodeEncodeError:
            pval = bytes(pval, "utf-8").decode(PGN_ENCODING, errors="ignore")
            print(f'[{tag} "{pval}"]', file=handle)
        processed_tags = processed_tags + [tag]

    # Mandatory ordered seven-tag roster
    status = reprResult[model.status]
    for tag in mandatory_tags:
        value = model.tags[tag]
        if tag == "Date":
            y, m, d = parseDateTag(value)
            y = "%04d" % y if y is not None else "????"
            m = "%02d" % m if m is not None else "??"
            d = "%02d" % d if d is not None else "??"
            value = f"{y}.{m}.{d}"
        elif value == "":
            value = "?"
        write_tag(tag, value, roster=True)
    write_tag("Result", reprResult[model.status], roster=True)

    # Variant
    if model.variant.variant != NORMALCHESS:
        write_tag("Variant", model.variant.cecp_name.capitalize())

    # Initial position
    if model.boards[0].asFen() != FEN_START:
        write_tag("SetUp", "1")
        write_tag("FEN", model.boards[0].asFen())

    # Number of moves
    write_tag("PlyCount", model.ply - model.lowply)

    # Final position
    if model.reason in ABORTED_REASONS:
        value = "abandoned"
    elif model.reason == WON_ADJUDICATION and model.isEngine2EngineGame():
        value = "rules infraction"
    elif model.reason in ADJUDICATION_REASONS:
        value = "adjudication"
    elif model.reason in DEATH_REASONS:
        value = "death"
    elif model.reason in CALLFLAG_REASONS:
        value = "time forfeit"
    elif model.reason in ADJOURNED_REASONS or status == "*":
        value = "unterminated"
    else:
        value = "normal"
    write_tag("Termination", value)

    # ELO and its variation
    if conf.get("saveRatingChange"):
        welo = model.tags["WhiteElo"]
        belo = model.tags["BlackElo"]
        if welo != "" and belo != "":
            write_tag(
                "WhiteRatingDiff", get_elo_rating_change_pgn(model, WHITE)
            )  # Unofficial
            write_tag(
                "BlackRatingDiff", get_elo_rating_change_pgn(model, BLACK)
            )  # Unofficial

    # Time
    if model.timed:
        write_tag(
            "WhiteClock",
            msToClockTimeTag(int(model.timemodel.getPlayerTime(WHITE) * 1000)),
        )
        write_tag(
            "BlackClock",
            msToClockTimeTag(int(model.timemodel.getPlayerTime(BLACK) * 1000)),
        )

    # Write all the unprocessed tags
    for tag in model.tags:
        # Debug: print(">> %s = %s" % (tag, str(model.tags[tag])))
        write_tag(tag, model.tags[tag])

    # Discovery of the moves and comments
    save_emt = conf.get("saveEmt")
    save_eval = conf.get("saveEval")
    result = []
    walk(model.boards[0].board, result, model, save_emt, save_eval)

    # Alignment of the fetched elements
    indented = conf.get("indentPgn")
    if indented:
        text_buffer = ""
        depth = 0
        crlf = False
        for text in result:
            # De/Indentation
            crlf = (text_buffer[-1:] if len(text_buffer) > 0 else "") in ["\r", "\n"]
            if text == "(":
                depth += 1
                if indented and not crlf:
                    text_buffer += os.linesep
                    crlf = True
            # Space between each term
            last = text_buffer[-1:] if len(text_buffer) > 0 else ""
            crlf = last in ["\r", "\n"]
            if (
                not crlf
                and last != " "
                and last != "\t"
                and last != "("
                and not text.startswith("\r")
                and not text.startswith("\n")
                and text != ")"
                and len(text_buffer) > 0
            ):
                text_buffer += " "
            # New line for a new main move
            if len(text_buffer) == 0 or (
                indented
                and depth == 0
                and last != "\r"
                and last != "\n"
                and re.match(r"^[0-9]+\.", text) is not None
            ):
                text_buffer += os.linesep
                crlf = True
            # Alignment
            if crlf and depth > 0:
                for j in range(0, depth):
                    text_buffer += "    "
            # Term
            text_buffer += text
            if indented and text == ")":
                text_buffer += os.linesep
                crlf = True
                depth -= 1
    else:
        # Add new line to separate tag section and movetext
        print("", file=handle)
        text_buffer = textwrap.fill(" ".join(result), width=80)

    # Final
    status = reprResult[model.status]
    try:
        print(text_buffer, status, file=handle)
    except UnicodeEncodeError:
        text_buffer = bytes(text_buffer, "utf-8").decode(PGN_ENCODING, errors="ignore")
        print(text_buffer, status, file=handle)
    # Add new line to separate next game
    print("", file=handle)

    output = handle.getvalue() if isinstance(handle, StringIO) else ""
    handle.close()
    return output


def walk(node, result, model, save_emt=False, save_eval=False, vari=False):
    """Prepares a game data for .pgn storage.
    Recursively walks the node tree to collect moves and comments
    into a resulting movetext string.

    Arguments:
    node - list (a tree of lboards created by the pgn parser)
    result - str (movetext strings)"""

    while True:
        if node is None:
            break

        # Initial game or variation comment
        if node.prev is None:
            for child in node.children:
                if isinstance(child, str):
                    result.append(f"{{{child}}}{os.linesep}")
            node = node.next
            continue

        movecount = move_count(
            node, black_periods=(save_emt or save_eval) and "TimeControl" in model.tags
        )
        if movecount is not None:
            if movecount:
                result.append(movecount)
            move = node.lastMove
            result.append(toSAN(node.prev, move))
            if (save_emt or save_eval) and not vari:
                emt_eval = ""
                if "TimeControl" in model.tags and save_emt:
                    elapsed = model.timemodel.getElapsedMoveTime(
                        node.plyCount - model.lowply
                    )
                    emt_eval = "[%%emt %s]" % formatTime(elapsed, clk2pgn=True)
                if node.plyCount in model.scores and save_eval:
                    moves, score, depth = model.scores[node.plyCount]
                    if node.color == BLACK:
                        score = -score
                    emt_eval += f"[%eval {score / 100.0:0.2f}/{depth}]"
                if emt_eval:
                    result.append("{%s}" % emt_eval)

        for nag in node.nags:
            if nag:
                result.append(nag)

        for child in node.children:
            if isinstance(child, str):
                # comment
                if child:
                    result.append("{%s}" % child)
            else:
                # variations
                if node.fen_was_applied:
                    result.append("(")
                    walk(child[0], result, model, save_emt, save_eval, vari=True)
                    result.append(")")
                    # variation after last played move is not valid pgn
                    # but we will save it as in comment
                else:
                    result.append("{%s:" % _("Analyzer's primary variation"))
                    walk(child[0], result, model, save_emt, save_eval, vari=True)
                    result.append("}")

        if node.next:
            node = node.next
        else:
            break


def move_count(node, black_periods=False):
    mvcount = None
    if node.fen_was_applied:
        ply = node.plyCount
        if ply % 2 == 1:
            mvcount = "%d." % (ply // 2 + 1)
        # initial game move, or initial variation move
        # it can be the same position as the main line! this is the reason using id()
        elif node.prev.prev is None or id(node) != id(node.prev.next) or black_periods:
            mvcount = "%d..." % (ply // 2)
        elif node.prev.children:
            # move after real(not [%foo bar]) comment
            need_mvcount = False
            for child in node.prev.children:
                if isinstance(child, str):
                    if not child.startswith("[%"):
                        need_mvcount = True
                        break
                else:
                    need_mvcount = True
                    break
            if need_mvcount:
                mvcount = "%d..." % (ply // 2)
            else:
                mvcount = ""
        else:
            mvcount = ""
    return mvcount


def load(handle, progressbar=None):
    return PGNFile(handle, progressbar)


cpuinfo = get_cpu()
MODERN = "-modern" if cpuinfo["popcnt"] else ""

altpath = getEngineDataPrefix()

if sys.platform == "win32":
    scoutfish = "scoutfish_x64.exe"
else:
    scoutfish = "scoutfish_x{}{}{}".format(
        cpuinfo["bitness"], MODERN, cpuinfo["binext"]
    )
scoutfish_path = shutil.which(scoutfish, mode=os.X_OK, path=altpath)

if sys.platform == "win32":
    parser = "parser_x64.exe"
else:
    parser = "parser_x{}{}{}".format(cpuinfo["bitness"], MODERN, cpuinfo["binext"])
chess_db_path = shutil.which(parser, mode=os.X_OK, path=altpath)


class PGNFile(ChessFile):
    def __init__(self, handle, progressbar=None):
        ChessFile.__init__(self, handle)
        self.handle = handle
        self.progressbar = progressbar
        self.pgn_is_string = isinstance(handle, StringIO)

        if self.pgn_is_string:
            self.games = [
                self.load_game_tags(),
            ]
        else:
            self.skip = 0
            self.limit = 100
            self.order_col = game.c.offset
            self.is_desc = False
            self.reset_last_seen()

            # filter expressions to .sqlite .bin .scout
            self.tag_query = None
            self.fen = None
            self.scout_query = None

            self.scoutfish = None
            self.chess_db = None

            self.sqlite_path = os.path.splitext(self.path)[0] + ".sqlite"
            self.engine = dbmodel.get_engine(self.sqlite_path)
            self.tag_database = TagDatabase(self.engine)

            self.games, self.offs_ply = self.get_records(0)
            log.info(
                f"{self.path} contains {self.count} game(s)",
                extra={"task": "SQL"},
            )

    def get_count(self):
        """Number of games in .pgn database"""
        if self.pgn_is_string:
            return len(self.games)
        else:
            return self.tag_database.count

    count = property(get_count)

    def get_size(self):
        """Size of .pgn file in bytes"""
        return os.path.getsize(self.path)

    size = property(get_size)

    def close(self):
        self.tag_database.close()
        ChessFile.close(self)

    def init_tag_database(self, importer=None):
        """Create/open .sqlite database of game header tags"""
        # Import .pgn header tags to .sqlite database

        sqlite_path = self.path.replace(".pgn", ".sqlite")
        if (
            os.path.isfile(self.path)
            and os.path.isfile(sqlite_path)
            and getmtime(self.path) > getmtime(sqlite_path)
        ):
            metadata.drop_all(self.engine)
            metadata.create_all(self.engine)
            ini_schema_version(self.engine)

        size = self.size
        if size > 0 and self.tag_database.count == 0:
            if size > 10000000:
                drop_indexes(self.engine)
            if self.progressbar is not None:
                from gi.repository import GLib

                GLib.idle_add(self.progressbar.set_text, _("Importing game headers..."))
            if importer is None:
                importer = PgnImport(self)
            importer.initialize()
            importer.do_import(self.path, progressbar=self.progressbar)
            if size > 10000000 and not importer.cancel:
                create_indexes(self.engine)

        return importer

    def init_chess_db(self):
        """Create/open polyglot .bin file with extra win/loss/draw stats
        using chess_db parser from https://github.com/mcostalba/chess_db
        """
        if chess_db_path is not None and self.path and self.size > 0:
            try:
                if self.progressbar is not None:
                    from gi.repository import GLib

                    GLib.idle_add(
                        self.progressbar.set_text, _("Creating .bin index file...")
                    )
                self.chess_db = Parser(engine=(chess_db_path,))
                self.chess_db.open(self.path)
                bin_path = os.path.splitext(self.path)[0] + ".bin"
                if not os.path.isfile(bin_path):
                    log.debug("No valid games found in %s" % self.path)
                    self.chess_db = None
                elif getmtime(self.path) > getmtime(bin_path):
                    self.chess_db.make()
            except OSError as err:
                self.chess_db = None
                log.warning(
                    "Failed to sart chess_db parser. OSError {} {}".format(
                        err.errno, err.strerror
                    )
                )
            except pexpect.TIMEOUT:
                self.chess_db = None
                log.warning("chess_db parser failed (pexpect.TIMEOUT)")
            except pexpect.EOF:
                self.chess_db = None
                log.warning("chess_db parser failed (pexpect.EOF)")

    def init_scoutfish(self):
        """Create/open .scout database index file to help querying
        using scoutfish from https://github.com/mcostalba/scoutfish
        """
        if scoutfish_path is not None and self.path and self.size > 0:
            try:
                if self.progressbar is not None:
                    from gi.repository import GLib

                    GLib.idle_add(
                        self.progressbar.set_text, _("Creating .scout index file...")
                    )
                self.scoutfish = Scoutfish(engine=(scoutfish_path,))
                self.scoutfish.open(self.path)
                scout_path = os.path.splitext(self.path)[0] + ".scout"
                if getmtime(self.path) > getmtime(scout_path):
                    self.scoutfish.make()
            except OSError as err:
                self.scoutfish = None
                log.warning(
                    "Failed to sart scoutfish. OSError {} {}".format(
                        err.errno, err.strerror
                    )
                )
            except pexpect.TIMEOUT:
                self.scoutfish = None
                log.warning("scoutfish failed (pexpect.TIMEOUT)")
            except pexpect.EOF:
                self.scoutfish = None
                log.warning("scoutfish failed (pexpect.EOF)")

    def get_book_moves(self, fen):
        """Get move-games-win-loss-draw stat of fen position"""
        rows = []
        if self.chess_db is not None:
            move_stat = self.chess_db.find(f"limit {1} skip {0} {fen}")
            for mstat in move_stat["moves"]:
                rows.append(
                    (
                        mstat["move"],
                        int(mstat["games"]),
                        int(mstat["wins"]),
                        int(mstat["losses"]),
                        int(mstat["draws"]),
                    )
                )
        return rows

    def has_position(self, fen):
        # ChessDB (prioritary)
        if self.chess_db is not None:
            ret = self.chess_db.find(f"limit {1} skip {0} {fen}")
            if len(ret["moves"]) > 0:
                return TOOL_CHESSDB, True
        # Scoutfish (alternate by approximation)
        if self.scoutfish is not None:
            q = {"limit": 1, "skip": 0, "sub-fen": fen}
            ret = self.scoutfish.scout(q)
            if ret["match count"] > 0:
                return TOOL_SCOUTFISH, True
        return TOOL_NONE, False

    def set_tag_order(self, order_col, is_desc):
        self.order_col = order_col
        self.is_desc = is_desc
        self.tag_database.build_order_by(self.order_col, self.is_desc)

    def reset_last_seen(self):
        col_max = "ZZZ" if isinstance(self.order_col.type, String) else 2**32
        col_min = "" if isinstance(self.order_col.type, String) else -1
        if self.is_desc:
            self.last_seen = [(col_max, 2**32)]
        else:
            self.last_seen = [(col_min, -1)]

    def set_tag_filter(self, query):
        """Set (now prefixing) text and
        create where clause we will use to query header tag .sqlite database
        """
        self.tag_query = query
        self.tag_database.build_where_tags(self.tag_query)

    def set_fen_filter(self, fen):
        """Set fen string we will use to get game offsets from .bin database"""
        if self.chess_db is not None and fen is not None and fen != FEN_START:
            self.fen = fen
        else:
            self.fen = None
            self.tag_database.build_where_offs8(None)

    def set_scout_filter(self, query):
        """Set json string we will use to get game offsets from  .scout database"""
        if self.scoutfish is not None and query:
            self.scout_query = query
        else:
            self.scout_query = None
            self.tag_database.build_where_offs(None)
            self.offs_ply = {}

    def get_offs(self, skip, filtered_offs_list=None):
        """Get offsets from .scout database and
        create where clause we will use to query header tag .sqlite database
        """
        if self.scout_query:
            limit = (10000 if self.tag_query else self.limit) + 1
            self.scout_query["skip"] = skip
            self.scout_query["limit"] = limit
            move_stat = self.scoutfish.scout(self.scout_query)

            offsets = []
            for mstat in move_stat["matches"]:
                offs = mstat["ofs"]
                if filtered_offs_list is None:
                    offsets.append(offs)
                    self.offs_ply[offs] = mstat["ply"][0]
                elif offs in filtered_offs_list:
                    offsets.append(offs)
                    self.offs_ply[offs] = mstat["ply"][0]

            if filtered_offs_list is not None:
                # Continue scouting until we get enough good offset if needed
                # print(0, move_stat["match count"], len(offsets))
                i = 1
                while len(offsets) < self.limit and move_stat["match count"] == limit:
                    self.scout_query["skip"] = i * limit - 1
                    move_stat = self.scoutfish.scout(self.scout_query)

                    for mstat in move_stat["matches"]:
                        offs = mstat["ofs"]
                        if offs in filtered_offs_list:
                            offsets.append(offs)
                            self.offs_ply[offs] = mstat["ply"][0]

                    # print(i, move_stat["match count"], len(offsets))
                    i += 1

            if len(offsets) > self.limit:
                self.tag_database.build_where_offs(offsets[: self.limit])
            else:
                self.tag_database.build_where_offs(offsets)

    def get_offs8(self, skip, filtered_offs_list=None):
        """Get offsets from .bin database and
        create where clause we will use to query header tag .sqlite database
        """
        if self.fen:
            move_stat = self.chess_db.find(f"limit {self.limit} skip {skip} {self.fen}")

            offsets = []
            for mstat in move_stat["moves"]:
                offs = mstat["pgn offsets"]
                if filtered_offs_list is None:
                    offsets += offs
                elif offs in filtered_offs_list:
                    offsets += offs

            if len(offsets) > self.limit:
                self.tag_database.build_where_offs8(sorted(offsets)[: self.limit])
            else:
                self.tag_database.build_where_offs8(sorted(offsets))

    def get_records(self, direction=FIRST_PAGE):
        """Get game header tag records from .sqlite database in paginated way"""
        if direction == FIRST_PAGE:
            self.skip = 0
            self.reset_last_seen()
        elif direction == NEXT_PAGE:
            if not self.tag_query:
                self.skip += self.limit
        elif direction == PREV_PAGE:
            if len(self.last_seen) == 2:
                self.reset_last_seen()
            elif len(self.last_seen) > 2:
                self.last_seen = self.last_seen[:-2]

            if not self.tag_query and self.skip >= self.limit:
                self.skip -= self.limit

        if self.fen:
            self.reset_last_seen()

        filtered_offs_list = None
        if self.tag_query and (self.fen or self.scout_query):
            filtered_offs_list = self.tag_database.get_offsets_for_tags(
                self.last_seen[-1]
            )

        if self.fen:
            self.get_offs8(self.skip, filtered_offs_list=filtered_offs_list)

        if self.scout_query:
            self.get_offs(self.skip, filtered_offs_list=filtered_offs_list)
            # No game satisfied scout_query
            if self.tag_database.where_offs is None:
                return [], {}

        records = self.tag_database.get_records(self.last_seen[-1], self.limit)

        if records:
            self.last_seen.append(
                (records[-1][col2label[self.order_col]], records[-1]["Offset"])
            )
            return records, self.offs_ply
        else:
            return [], {}

    def load_game_tags(self):
        """Reads header tags from pgn if pgn is a one game only StringIO object"""

        header = collections.defaultdict(str)
        header["Id"] = 0
        header["Offset"] = 0
        for line in self.handle.readlines():
            line = line.strip()
            if line.startswith("[") and line.endswith("]"):
                tag_match = TAG_REGEX.match(line)
                if tag_match:
                    value = tag_match.group(2)
                    value = value.replace('\\"', '"')
                    value = value.replace("\\\\", "\\")
                    header[tag_match.group(1)] = value
            else:
                break
        return header

    def loadToModel(self, rec, position=-1, model=None):
        """Parse game text and load game record header tags to a GameModel object"""

        if model is None:
            model = GameModel()

        if self.pgn_is_string:
            rec = self.games[0]

        # Load mandatory tags
        for tag in mandatory_tags:
            model.tags[tag] = rec[tag]

        # Load other tags
        for tag in ("WhiteElo", "BlackElo", "ECO", "TimeControl", "Annotator"):
            model.tags[tag] = rec[tag]

        if self.pgn_is_string:
            for tag in rec:
                if isinstance(rec[tag], str) and rec[tag]:
                    model.tags[tag] = rec[tag]
        else:
            model.info = self.tag_database.get_info(rec)
            extra_tags = self.tag_database.get_exta_tags(rec)
            for et in extra_tags:
                model.tags[et["tag_name"]] = et["tag_value"]

        if self.pgn_is_string:
            variant = rec["Variant"].capitalize()
        else:
            variant = self.get_variant(rec)

        if model.tags["TimeControl"]:
            tc = parseTimeControlTag(model.tags["TimeControl"])
            if tc is not None:
                secs, gain, moves = tc
                model.timed = True
                model.timemodel.secs = secs
                model.timemodel.gain = gain
                model.timemodel.minutes = secs / 60
                model.timemodel.moves = moves
                for tag, color in (("WhiteClock", WHITE), ("BlackClock", BLACK)):
                    if tag in model.tags:
                        try:
                            millisec = parseClockTimeTag(model.tags[tag])
                            # We need to fix when FICS reports negative clock time like this
                            # [TimeControl "180+0"]
                            # [WhiteClock "0:00:15.867"]
                            # [BlackClock "23:59:58.820"]
                            start_sec = (
                                (millisec - 24 * 60 * 60 * 1000) / 1000.0
                                if millisec > 23 * 60 * 60 * 1000
                                else millisec / 1000.0
                            )
                            model.timemodel.intervals[color][0] = start_sec
                        except ValueError:
                            raise LoadingError("Error parsing '%s'" % tag)
        fenstr = rec["FEN"]

        if variant:
            if variant not in name2variant:
                raise LoadingError("Unknown variant %s" % variant)

            model.tags["Variant"] = variant
            # Fixes for some non statndard Chess960 .pgn
            if (fenstr is not None) and variant == "Fischerandom":
                parts = fenstr.split()
                parts[0] = parts[0].replace(".", "/").replace("0", "")
                if len(parts) == 1:
                    parts.append("w")
                    parts.append("-")
                    parts.append("-")
                fenstr = " ".join(parts)

            model.variant = name2variant[variant]
            board = LBoard(model.variant.variant)
        else:
            model.variant = NormalBoard
            board = LBoard()

        if fenstr:
            try:
                board.applyFen(fenstr)
                model.tags["FEN"] = fenstr
            except SyntaxError as err:
                board.applyFen(FEN_EMPTY)
                raise LoadingError(
                    _("The game can't be loaded, because of an error parsing FEN"),
                    err.args[0],
                )
        else:
            board.applyFen(FEN_START)

        boards = [board]

        del model.moves[:]
        del model.variations[:]

        self.error = None
        movetext = self.get_movetext(rec)
        boards = self.parse_movetext(movetext, boards[0], position)

        # The parser built a tree of lboard objects, now we have to
        # create the high level Board and Move lists...

        for board in boards:
            if board.lastMove is not None:
                model.moves.append(Move(board.lastMove))

        self.has_emt = False
        self.has_eval = False

        def _create_board(model, node):
            if node.prev is None:
                # initial game board
                board = model.variant(setup=node.asFen(), lboard=node)
            else:
                move = Move(node.lastMove)
                try:
                    board = node.prev.pieceBoard.move(move, lboard=node)
                except Exception:
                    raise LoadingError(
                        _("Invalid move."),
                        f"{move_count(node, black_periods=True)}{move}",
                    )

            return board

        def walk(model, node, path):
            boards = path
            stack = []
            current = node

            while current is not None:
                board = _create_board(model, current)
                boards.append(board)
                stack.append(current)
                current = current.next
            else:
                model.variations.append(list(boards))

            while stack:
                current = stack.pop()
                boards.pop()
                for child in current.children:
                    if isinstance(child, list):
                        if len(child) > 1:
                            # non empty variation, go walk
                            walk(model, child[1], list(boards))
                    else:
                        if not self.has_emt:
                            self.has_emt = child.find("%emt") >= 0
                        if not self.has_eval:
                            self.has_eval = child.find("%eval") >= 0

        # Collect all variation paths into a list of board lists
        # where the first one will be the boards of mainline game.
        # model.boards will allways point to the current shown variation
        # which will be model.variations[0] when we are in the mainline.
        walk(model, boards[0], [])
        model.boards = model.variations[0]
        self.has_emt = self.has_emt and model.timed
        if self.has_emt or self.has_eval:
            if self.has_emt:
                blacks = len(model.moves) // 2
                whites = len(model.moves) - blacks

                model.timemodel.intervals = [
                    [model.timemodel.intervals[0][0]] * (whites + 1),
                    [model.timemodel.intervals[1][0]] * (blacks + 1),
                ]
                model.timemodel.intervals[0][0] = secs
                model.timemodel.intervals[1][0] = secs
            for ply, board in enumerate(boards):
                for child in board.children:
                    if isinstance(child, str):
                        if self.has_emt:
                            match = move_time_re.search(child)
                            if match:
                                movecount, color = divmod(ply + 1, 2)
                                hour, minute, sec, msec = match.groups()
                                prev = model.timemodel.intervals[color][movecount - 1]
                                hour = 0 if hour is None else int(hour[:-1])
                                minute = 0 if minute is None else int(minute[:-1])
                                msec = 0 if msec is None else int(msec)
                                msec += (
                                    int(sec) * 1000
                                    + int(minute) * 60 * 1000
                                    + int(hour) * 60 * 60 * 1000
                                )
                                model.timemodel.intervals[color][movecount] = (
                                    prev - msec / 1000.0 + gain
                                )

                        if self.has_eval:
                            match = move_eval_re.search(child)
                            if match:
                                sign, num, fraction, depth = match.groups()
                                sign = 1 if sign is None or sign == "+" else -1
                                num = int(num)
                                fraction = 0 if fraction is None else int(fraction)
                                value = sign * (num * 100 + fraction)
                                depth = "" if depth is None else depth
                                if board.color == BLACK:
                                    value = -value
                                model.scores[ply] = ("", value, depth)
            log.debug("pgn.loadToModel: intervals %s" % model.timemodel.intervals)

        # Find the physical status of the game
        model.status, model.reason = getStatus(model.boards[-1])

        # Apply result from .pgn if the last position was loaded
        if position == -1 or len(model.moves) == position - model.lowply:
            if self.pgn_is_string:
                result = rec["Result"]
                if result in pgn2Const:
                    status = pgn2Const[result]
                else:
                    status = RUNNING
            else:
                status = rec["Result"]

            if status in (WHITEWON, BLACKWON) and status != model.status:
                model.status = status
                model.reason = WON_RESIGN
            elif status == DRAW and status != model.status:
                model.status = DRAW
                model.reason = DRAW_AGREE

        if model.timed:
            model.timemodel.movingColor = model.boards[-1].color

        # If parsing gave an error we throw it now, to enlarge our possibility
        # of being able to continue the game from where it failed.
        if self.error:
            raise self.error

        return model

    def parse_movetext(self, string, board, position, variation=False):
        """Recursive parses a movelist part of one game.

        Arguments:
        srting - str (movelist)
        board - lboard (initial position)
        position - int (maximum ply to parse)
        variation- boolean (True if the string is a variation)"""

        boards = []
        boards_append = boards.append

        last_board = board
        if variation:
            # this board used only to hold initial variation comments
            boards_append(LBoard(board.variant))
        else:
            # initial game board
            boards_append(board)

        # status = None
        parenthesis = 0
        v_string = ""
        v_last_board = None
        for m in re.finditer(pattern, string):
            group, text = m.lastindex, m.group(m.lastindex)
            if parenthesis > 0:
                v_string += " " + text

            if group == VARIATION_END:
                parenthesis -= 1
                if parenthesis == 0:
                    if last_board.prev is None:
                        errstr1 = _("Error parsing %(mstr)s") % {"mstr": string}
                        self.error = LoadingError(errstr1, "")
                        return boards  # , status

                    v_last_board.children.append(
                        self.parse_movetext(
                            v_string[:-1], last_board.prev, position, variation=True
                        )
                    )
                    v_string = ""
                    continue

            elif group == VARIATION_START:
                parenthesis += 1
                if parenthesis == 1:
                    v_last_board = last_board

            if parenthesis == 0:
                if group == FULL_MOVE:
                    if not variation:
                        if position != -1 and last_board.plyCount >= position:
                            break

                    mstr = m.group(MOVE)
                    try:
                        lmove = parseAny(last_board, mstr)
                    except ParsingError as err:
                        # TODO: save the rest as comment
                        # last_board.children.append(string[m.start():])
                        notation, reason, boardfen = err.args
                        ply = last_board.plyCount
                        if ply % 2 == 0:
                            moveno = "%d." % (ply // 2 + 1)
                        else:
                            moveno = "%d..." % (ply // 2 + 1)
                        errstr1 = _(
                            "The game can't be read to end, because of an error parsing move %(moveno)s '%(notation)s'."
                        ) % {"moveno": moveno, "notation": notation}
                        errstr2 = _("The move failed because %s.") % reason
                        self.error = LoadingError(errstr1, errstr2)
                        break
                    except Exception:
                        ply = last_board.plyCount
                        if ply % 2 == 0:
                            moveno = "%d." % (ply // 2 + 1)
                        else:
                            moveno = "%d..." % (ply // 2 + 1)
                        errstr1 = _("Error parsing move %(moveno)s %(mstr)s") % {
                            "moveno": moveno,
                            "mstr": mstr,
                        }
                        self.error = LoadingError(errstr1, "")
                        break

                    new_board = last_board.clone()
                    new_board.applyMove(lmove)

                    if m.group(MOVE_COMMENT):
                        new_board.nags.append(symbol2nag(m.group(MOVE_COMMENT)))

                    new_board.prev = last_board

                    # set last_board next, except starting a new variation
                    if variation and last_board == board:
                        boards[0].next = new_board
                    else:
                        last_board.next = new_board

                    boards_append(new_board)
                    last_board = new_board

                elif group == COMMENT_REST:
                    last_board.children.append(text[1:])

                elif group == COMMENT_BRACE:
                    comm = text.replace("{\r\n", "{").replace("\r\n}", "}")
                    # Preserve new lines of lichess study comments
                    if self.path is not None and "lichess_study_" in self.path:
                        comment = comm[1:-1]
                    else:
                        comm = comm[1:-1].splitlines()
                        comment = " ".join([line.strip() for line in comm])
                    if variation and last_board == board:
                        # initial variation comment
                        boards[0].children.append(comment)
                    else:
                        last_board.children.append(comment)

                elif group == COMMENT_NAG:
                    last_board.nags.append(text)

                # TODO
                elif group == RESULT:
                    # if text == "1/2":
                    #    status = reprResult.index("1/2-1/2")
                    # else:
                    #    status = reprResult.index(text)
                    break

                else:
                    print("Unknown:", text)

        return boards  # , status

    def get_movetext(self, rec):
        self.handle.seek(rec["Offset"])
        in_comment = False
        lines = []
        line = self.handle.readline()
        if not line.strip():
            line = self.handle.readline()

        while line:
            # escape non-PGN data line
            if line.startswith("%"):
                line = self.handle.readline()
                continue

            # header tag line
            if not in_comment and line.startswith("["):
                line = self.handle.readline()
                continue

            # update in_comment state
            if (not in_comment and "{" in line) or (in_comment and "}" in line):
                in_comment = line.rfind("{") > line.rfind("}")

            # if there is something add it
            if line.strip():
                if not self.pgn_is_string and self.handle.pgn_encoding != PGN_ENCODING:
                    line = line.encode(PGN_ENCODING).decode(self.handle.pgn_encoding)
                lines.append(line)
                line = self.handle.readline()
            # if line is empty it should be the game separator line except...
            elif len(lines) == 0 or in_comment:
                if in_comment:
                    lines.append(line)
                line = self.handle.readline()
            else:
                break
        return "".join(lines)

    def get_variant(self, rec):
        variant = rec["Variant"]
        return variants[variant].cecp_name.capitalize() if variant else ""


nag2symbolDict = {
    "$0": "",
    "$1": "!",
    "$2": "?",
    "$3": "!!",
    "$4": "??",
    "$5": "!?",
    "$6": "?!",
    "$7": "□",  # forced move
    "$8": "□",
    "$9": "??",
    "$10": "=",
    "$11": "=",
    "$12": "=",
    "$13": "∞",  # unclear
    "$14": "+=",
    "$15": "=+",
    "$16": "±",
    "$17": "∓",
    "$18": "+-",
    "$19": "-+",
    "$20": "+--",
    "$21": "--+",
    "$22": "⨀",  # zugzwang
    "$23": "⨀",
    "$24": "◯",  # space
    "$25": "◯",
    "$26": "◯",
    "$27": "◯",
    "$28": "◯",
    "$29": "◯",
    "$32": "⟳",  # development
    "$33": "⟳",
    "$36": "↑",  # initiative
    "$37": "↑",
    "$40": "→",  # attack
    "$41": "→",
    "$44": "~=",  # compensation
    "$45": "=~",
    "$132": "⇆",  # counterplay
    "$133": "⇆",
    "$136": "⨁",  # time
    "$137": "⨁",
    "$138": "⨁",
    "$139": "⨁",
    "$140": "∆",  # with the idea
    "$141": "∇",  # aimed against
    "$142": "⌓",  # better is
    "$146": "N",  # novelty
}

symbol2nagDict = {}
for k, v in nag2symbolDict.items():
    if v not in symbol2nagDict:
        symbol2nagDict[v] = k


def nag2symbol(nag):
    return nag2symbolDict.get(nag, nag)


def symbol2nag(symbol):
    return symbol2nagDict[symbol]



"""Shuffle Variant"""

import random

from pychess.Utils.const import SHUFFLECHESS, VARIANTS_SHUFFLE
from pychess.Utils.Board import Board


class ShuffleBoard(Board):
    """:Description: The shuffle variant uses the standard chess rules with the exception
    no castling is allowed and the back rank is shuffled around
    """

    variant = SHUFFLECHESS
    __desc__ = _(
        "xboard nocastle: http://home.hccnet.nl/h.g.muller/engine-intf.html#8\n"
        + "FICS wild/2: http://www.freechess.org/Help/HelpFiles/wild.html\n"
        + "* Random arrangement of the pieces behind the pawns\n"
        + "* No castling\n"
        + "* Black's arrangement mirrors white's"
    )
    name = _("Shuffle")
    cecp_name = "nocastle"
    need_initial_board = True
    standard_rules = True
    variant_group = VARIANTS_SHUFFLE

    def __init__(self, setup=False, lboard=None):
        if setup is True:
            Board.__init__(self, setup=self.shuffle_start(), lboard=lboard)
        else:
            Board.__init__(self, setup=setup, lboard=lboard)

    def shuffle_start(self):
        back_rank = ["r", "n", "b", "q", "k", "b", "n", "r"]
        random.shuffle(back_rank)
        fen = "".join(back_rank)
        fen = fen + "/pppppppp/8/8/8/8/PPPPPPPP/" + fen.upper() + " w - - 0 1"
        return fen


if __name__ == "__main__":
    Board = ShuffleBoard(True)
    for i in range(10):
        print(Board.shuffle_start())



from array import array


# setBit returns a bitboard with the ith bit set
def setBit(bitboard, i):
    return bitboard | bitPosArray[i]


bitPosArray = [2 ** (63 - i) for i in range(64)]


# clearBit returns the bitboard with the ith bit unset
def clearBit(bitboard, i):
    return bitboard & notBitPosArray[i]


notBitPosArray = [~(2 ** (63 - i)) for i in range(64)]


# firstBit returns the bit closest to 0 (A4) that is set in the board
def firstBit(bitboard):
    """Returns the index of the first non-zero bit from left"""
    if bitboard >> 48:
        return lzArray[bitboard >> 48]
    if bitboard >> 32:
        return lzArray[bitboard >> 32] + 16
    if bitboard >> 16:
        return lzArray[bitboard >> 16] + 32
    return lzArray[bitboard] + 48


# The bitCount array returns the leading non-zero bit in the 16 bit
# input argument.

lzArray = array("B", [0] * 65536)

s = n = 1
for i in range(16):
    for j in range(s, s + n):
        lzArray[j] = 16 - 1 - i
    s += n
    n += n


# lastBit returns the bit closest to 63 (H8) that is set in the board
def lastBit(bitboard):
    return lsb[bitboard & -bitboard]


lsb = {}
for i in range(64):
    lsb[2**i] = 63 - i


# iterBits yields, or returns a list of, the positions of all set bits in a
# bitboard. There is no guarantee of the order.
def iterBits(bitboard):
    while bitboard:
        bit = bitboard & -bitboard
        yield lsb[bit]
        bitboard -= bit


# toString returns a representation of the bitboard for debugging
def toString(bitboard):
    s = []
    last = -1

    while bitboard:
        cord = firstBit(bitboard)
        bitboard = clearBit(bitboard, cord)
        for c in range(cord - last - 1):
            s.append(" -")
        s.append(" #")
        last = cord
    while len(s) < 64:
        s.append(" -")

    s2 = ""
    for i in range(64, 0, -8):
        a = s[i - 8 : i]
        s2 += "".join(a) + "\n"
    return s2



from gi.repository import Gtk
from pychess.System import uistuff
from pychess.System.prefix import addDataPrefix
from pychess.Utils.const import reprCord
from pychess.Utils.repr import reprColor, reprPiece
from pychess.Utils.lutils.lmove import TCORD
from pychess.Utils.lutils.leval import evalMaterial
from pychess.Utils.lutils import strateval

__title__ = _("Comments")

__icon__ = addDataPrefix("glade/panel_comments.svg")

__desc__ = _("The comments panel will try to analyze and explain the moves played")


class Sidepanel:
    def __init__(self):
        self.givenTips = {}

    def load(self, gmwidg):
        self.gamemodel = gmwidg.board.view.model
        self.model_cids = [
            self.gamemodel.connect_after("game_changed", self.game_changed),
            self.gamemodel.connect_after("game_started", self.game_started),
            self.gamemodel.connect_after("moves_undone", self.moves_undone),
            self.gamemodel.connect_after("game_terminated", self.on_game_terminated),
        ]

        scrollwin = Gtk.ScrolledWindow()
        self.tv = Gtk.TreeView()
        self.tv.set_headers_visible(False)
        scrollwin.add(self.tv)
        scrollwin.show_all()

        self.store = Gtk.ListStore(str)
        self.tv.set_model(self.store)
        self.tv.get_selection().set_mode(Gtk.SelectionMode.BROWSE)
        uistuff.appendAutowrapColumn(self.tv, "Comment", text=0)
        self.tv_cid = self.tv.connect("cursor_changed", self.cursorChanged)

        self.boardview = gmwidg.board.view
        self.cid = self.boardview.connect("shownChanged", self.shownChanged)

        return scrollwin

    def on_game_terminated(self, model):
        self.tv.disconnect(self.tv_cid)
        for cid in self.model_cids:
            self.gamemodel.disconnect(cid)
        self.boardview.disconnect(self.cid)

    def cursorChanged(self, tv):
        path, focus_column = tv.get_cursor()
        indices = path.get_indices()
        if indices:
            row = indices[0]
            board = self.gamemodel.boards[row]
            self.boardview.setShownBoard(board)

    def shownChanged(self, boardview, shown):
        if self.gamemodel is None:
            return
        if not boardview.shownIsMainLine():
            return
        row = shown - self.gamemodel.lowply

        try:
            iter = self.store.get_iter(row)
            selection = self.tv.get_selection()
            if selection is not None:
                selection.select_iter(iter)
                self.tv.scroll_to_cell(row)
        except ValueError:
            pass
            # deleted variations by moves_undoing

    def moves_undone(self, game, moves):
        model = self.tv.get_model()
        for i in range(moves):
            model.remove(model.get_iter((len(model) - 1,)))

    def game_started(self, model):
        if model.lesson_game:
            return
        self.game_changed(model, model.ply)

    def game_changed(self, model, ply):
        if len(self.store) == 0:
            for i in range(len(self.store) + model.lowply, ply + 1):
                self.addComment(model, self.__chooseComment(model, i))
        else:
            self.addComment(model, self.__chooseComment(model, ply))

        self.shownChanged(self.boardview, ply)

    def addComment(self, model, comment):
        if self.gamemodel is None or self.tv is None:
            return
        self.store.append([comment])

        # If latest ply is shown, we select the new latest
        selection = self.tv.get_selection()
        if selection is None:
            return
        iter = selection.get_selected()[1]
        if iter:
            path = self.tv.get_model().get_path(iter)
            indices = path.get_indices()
            if indices:
                row = indices[0]
                if row < self.boardview.shown - 1:
                    return

        if self.boardview.shown >= model.ply:
            iter = self.store.get_iter(len(self.store) - 1)
            self.tv.get_selection().select_iter(iter)

    def __chooseComment(self, model, ply):
        if ply == model.lowply:
            return _("Initial position")

        ########################################################################
        # Set up variables
        ########################################################################

        color = model.getBoardAtPly(ply - 1).board.color
        s, phase = evalMaterial(
            model.getBoardAtPly(ply).board, model.getBoardAtPly(ply - 1).color
        )

        #   * Final: Will be shown alone: "mates", "draws"
        #   * Prefix: Will always be shown: "castles", "promotes"
        #   * Attack: Will always be shown: "threaten", "preassures", "defendes"
        #   * Moves (s): Will always be shown: "put into *"
        #   * State: (s) Will always be shown: "new *"
        #   * Simple: (s) Max one will be shown: "develops", "activity"
        #   * Tip: (s) Will sometimes be shown: "pawn storm", "cramped position"

        ########################################################################
        # Call strategic evaluation functions
        ########################################################################

        def getMessages(prefix):
            messages = []
            for functionName in dir(strateval):
                if not functionName.startswith(prefix + "_"):
                    continue
                function = getattr(strateval, functionName)
                messages.extend(function(model, ply, phase))
            return messages

        # move = model.moves[-1].move
        # print "----- %d - %s -----" % (model.ply/2, toSAN(oldboard, move))

        # ----------------------------------------------------------------------
        # Final
        # ----------------------------------------------------------------------

        messages = getMessages("final")
        if messages:
            return f"{reprColor[color]} {messages[0]}"

        # ---

        strings = []

        # ----------------------------------------------------------------------
        # Attacks
        # ----------------------------------------------------------------------

        messages = getMessages("attack")
        for message in messages:
            strings.append(f"{reprColor[color]} {message}")

        # ----------------------------------------------------------------------
        # Check for prefixes
        # ----------------------------------------------------------------------

        messages = getMessages("prefix")
        if messages:
            prefix = messages[0]
        else:
            prefix = ""

        # ----------------------------------------------------------------------
        # Check for special move stuff. All of which accept prefixes
        # ----------------------------------------------------------------------

        for message in getMessages("offencive_moves") + getMessages("defencive_moves"):
            if prefix:
                strings.append(
                    "{} {} {} {}".format(
                        reprColor[color], prefix, _("and") + "\n", message
                    )
                )
                prefix = ""
            else:
                strings.append(f"{reprColor[color]} {message}")

        # ----------------------------------------------------------------------
        # Simple
        # ----------------------------------------------------------------------

        # We only add simples if there hasn't been too much stuff to say
        if not strings:
            messages = getMessages("simple")
            if messages:
                messages.sort(reverse=True)
                score, message = messages[0]
                if prefix:
                    strings.append(
                        "%s %s %s %s"
                        % (reprColor[color], prefix, _("and") + "\n", message)
                    )
                    prefix = ""

        # ----------------------------------------------------------------------
        # Prefix fallback
        # ----------------------------------------------------------------------

        # There was nothing to apply the prefix to, so we just post it here
        # before the states and tips
        if prefix:
            strings.append(f"{reprColor[color]} {prefix}")
            prefix = ""

        # ----------------------------------------------------------------------
        # State
        # ----------------------------------------------------------------------

        messages = getMessages("state")
        messages.sort(reverse=True)
        for score, message in messages:
            strings.append(message)

        # ----------------------------------------------------------------------
        # Tips
        # ----------------------------------------------------------------------

        tips = getMessages("tip")
        tips.sort(reverse=True)

        for score, tip in tips:
            if tip in self.givenTips:
                oldscore, oldply = self.givenTips[tip]
                if score < oldscore * 1.3 or model.ply < oldply + 10:
                    continue

            self.givenTips[tip] = (score, model.ply)
            strings.append(tip)
            break

        # ----------------------------------------------------------------------
        # Last solution
        # ----------------------------------------------------------------------

        if not strings:
            tcord = TCORD(model.getMoveAtPly(ply - 1).move)
            piece = model.getBoardAtPly(ply).board.arBoard[tcord]
            strings.append(
                _("%(color)s moves a %(piece)s to %(cord)s")
                % {
                    "color": reprColor[color],
                    "piece": reprPiece[piece],
                    "cord": reprCord[tcord],
                }
            )

        return ";\n".join(strings)



from gi.repository import GObject

from pychess.ic.managers.AutoLogOutManager import AutoLogOutManager


class ICCAutoLogOutManager(AutoLogOutManager):
    def __init__(self, connection):
        GObject.GObject.__init__(self)
        self.connection = connection



import os
from xml.dom import minidom
from collections import defaultdict


from pychess.System.prefix import addDataPrefix

from .PyDockLeaf import PyDockLeaf
from .PyDockComposite import PyDockComposite
from .ArrowButton import ArrowButton
from .HighlightArea import HighlightArea
from .__init__ import TabReceiver
from .__init__ import NORTH, EAST, SOUTH, WEST, CENTER


class PyDockTop(PyDockComposite, TabReceiver):
    def __init__(self, id, perspective):
        TabReceiver.__init__(self, perspective)
        self.id = id
        self.perspective = perspective
        self.set_no_show_all(True)
        self.highlightArea = HighlightArea(self)

        self.button_cids = defaultdict(list)

        self.buttons = (
            ArrowButton(self, addDataPrefix("glade/dock_top.svg"), NORTH),
            ArrowButton(self, addDataPrefix("glade/dock_right.svg"), EAST),
            ArrowButton(self, addDataPrefix("glade/dock_bottom.svg"), SOUTH),
            ArrowButton(self, addDataPrefix("glade/dock_left.svg"), WEST),
        )

        for button in self.buttons:
            self.button_cids[button] += [
                button.connect("dropped", self.__onDrop),
                button.connect("hovered", self.__onHover),
                button.connect("left", self.__onLeave),
            ]

    def _del(self):
        if self.highlightArea.handler_is_connected(self.highlightArea.cid):
            self.highlightArea.disconnect(self.highlightArea.cid)

        for button in self.buttons:
            for cid in self.button_cids[button]:
                if button.handler_is_connected(cid):
                    button.disconnect(cid)
            button.myparent = None

        self.button_cids = {}
        self.highlightArea.myparent = None

        TabReceiver._del(self)
        PyDockComposite._del(self)

    def getPosition(self):
        return CENTER

    def __repr__(self):
        return "top (%s)" % self.id

    # ===========================================================================
    #    Component stuff
    # ===========================================================================

    def addComponent(self, widget):
        self.add(widget)
        widget.show()

    def changeComponent(self, old, new):
        self.removeComponent(old)
        self.addComponent(new)

    def removeComponent(self, widget):
        self.remove(widget)

    def getComponents(self):
        child = self.get_child()
        if isinstance(child, PyDockComposite) or isinstance(child, PyDockLeaf):
            return [child]
        return []

    def dock(self, widget, position, title, id):
        if not self.getComponents():
            leaf = PyDockLeaf(widget, title, id, self.perspective)
            self.addComponent(leaf)
            return leaf
        else:
            return self.get_child().dock(widget, position, title, id)

    def clear(self):
        self.remove(self.get_child())

    # ===========================================================================
    #    Signals
    # ===========================================================================

    def showArrows(self):
        for button in self.buttons:
            button._calcSize()
            button.show()

    def hideArrows(self):
        for button in self.buttons:
            button.hide()
        self.highlightArea.hide()

    def __onDrop(self, arrowButton, sender):
        self.highlightArea.hide()
        child = sender.get_nth_page(sender.get_current_page())

        for instance in sender.get_parent().getInstances(self.perspective):
            instance.hideArrows()

        title, id = sender.get_parent().undock(child)
        self.dock(child, arrowButton.myposition, title, id)

    def __onHover(self, arrowButton, widget):
        self.highlightArea.showAt(arrowButton.myposition)
        arrowButton.get_window().raise_()

    def __onLeave(self, arrowButton):
        self.highlightArea.hide()

    # ===========================================================================
    #    XML
    # ===========================================================================

    def saveToXML(self, xmlpath):
        """
        <docks>
            <dock id="x">
                <v pos="200">
                    <leaf current="x" dockable="False">
                        <panel id="x" />
                    </leaf>
                    <h pos="200">
                        <leaf current="y" dockable="True">
                            <panel id="y" />
                            <panel id="z" />
                        </leaf>
                        <leaf current="y" dockable="True">
                            <panel id="y" />
                        </leaf>
                    </h>
                </v>
            </dock>
        </docks>
        """
        dockElem = None

        if os.path.isfile(xmlpath):
            doc = minidom.parse(xmlpath)
            for elem in doc.getElementsByTagName("dock"):
                if elem.getAttribute("id") == self.id:
                    for node in elem.childNodes:
                        elem.removeChild(node)
                    dockElem = elem
                    break

        if not dockElem:
            doc = minidom.getDOMImplementation().createDocument(None, "docks", None)
            dockElem = doc.createElement("dock")
            dockElem.setAttribute("id", self.id)
            doc.documentElement.appendChild(dockElem)

        if self.get_child():
            self.__addToXML(self.get_child(), dockElem, doc)
        with open(xmlpath, "w") as f_handle:
            doc.writexml(f_handle)
        doc.unlink()

    def __addToXML(self, component, parentElement, document):
        if isinstance(component, PyDockComposite):
            pos = component.paned.get_position()
            if component.getPosition() in (NORTH, SOUTH):
                childElement = document.createElement("v")
                size = float(component.get_allocation().height)
            else:
                childElement = document.createElement("h")
                size = float(component.get_allocation().width)
            #             if component.getPosition() in (NORTH, SOUTH):
            #                 print "saving v position as %s out of %s (%s)" % (str(pos), str(size), str(pos/max(size,pos)))
            childElement.setAttribute("pos", str(pos / max(size, pos)))
            self.__addToXML(component.getComponents()[0], childElement, document)
            self.__addToXML(component.getComponents()[1], childElement, document)

        elif isinstance(component, PyDockLeaf):
            childElement = document.createElement("leaf")
            childElement.setAttribute("current", component.getCurrentPanel())
            childElement.setAttribute("dockable", str(component.isDockable()))
            for panel, title, id in component.getPanels():
                element = document.createElement("panel")
                element.setAttribute("id", id)
                element.setAttribute("visible", str(panel.get_visible()))
                childElement.appendChild(element)

        parentElement.appendChild(childElement)

    def loadFromXML(self, xmlpath, idToWidget):
        """idTowidget is a dictionary {id: (widget,title)}
        asserts that self.id is in the xmlfile"""
        doc = minidom.parse(xmlpath)
        for elem in doc.getElementsByTagName("dock"):
            if elem.getAttribute("id") == self.id:
                break
        else:
            raise AttributeError(
                "XML file contains no <dock> elements with id '%s'" % self.id
            )

        child = [n for n in elem.childNodes if isinstance(n, minidom.Element)]
        if child:
            self.addComponent(self.__createWidgetFromXML(child[0], idToWidget))

    def __createWidgetFromXML(self, parentElement, idToWidget):
        children = [
            n for n in parentElement.childNodes if isinstance(n, minidom.Element)
        ]

        if parentElement.tagName in ("h", "v"):
            child1, child2 = children
            if parentElement.tagName == "h":
                new = PyDockComposite(EAST, self.perspective)
            else:
                new = PyDockComposite(SOUTH, self.perspective)
            new.initChildren(
                self.__createWidgetFromXML(child1, idToWidget),
                self.__createWidgetFromXML(child2, idToWidget),
                preserve_dimensions=True,
            )

            def cb(widget, event, pos):
                allocation = widget.get_allocation()
                if parentElement.tagName == "h":
                    widget.set_position(int(allocation.width * pos))
                else:
                    # print "loading v position as %s out of %s (%s)" % \
                    # (int(allocation.height * pos), str(allocation.height), str(pos))
                    widget.set_position(int(allocation.height * pos))
                widget.disconnect(conid)

            conid = new.paned.connect(
                "size-allocate", cb, float(parentElement.getAttribute("pos"))
            )
            return new

        elif parentElement.tagName == "leaf":
            id = children[0].getAttribute("id")
            try:
                title, widget, menu_item = idToWidget[id]
            except KeyError:
                id = self.old2new(id)
                title, widget, menu_item = idToWidget[id]

            leaf = PyDockLeaf(widget, title, id, self.perspective)
            visible = children[0].getAttribute("visible")
            visible = visible == "" or visible == "True"
            widget.set_visible(visible)
            if menu_item is not None:
                menu_item.set_active(visible)

            for panelElement in children[1:]:
                id = panelElement.getAttribute("id")
                try:
                    title, widget, menu_item = idToWidget[id]
                except KeyError:
                    id = self.old2new(id)
                    title, widget, menu_item = idToWidget[id]

                visible = panelElement.getAttribute("visible")
                visible = visible == "" or visible == "True"
                leaf.dock(widget, CENTER, title, id)
                widget.set_visible(visible)
                if menu_item is not None:
                    menu_item.set_active(visible)

            leaf.setCurrentPanel(self.old2new(parentElement.getAttribute("current")))
            if parentElement.getAttribute("dockable").lower() == "false":
                leaf.setDockable(False)
            return leaf

    def old2new(self, name):
        """After 0.99.0 database perspective panel names changed"""
        x = {
            "switcher": "SwitcherPanel",
            "openingtree": "OpeningTreePanel",
            "filter": "FilterPanel",
            "preview": "PreviewPanel",
            "chat": "ChatPanel",
            "console": "ConsolePanel",
            "news": "NewsPanel",
            "seeklist": "SeekListPanel",
            "seekgraph": "SeekGraphPanel",
            "playerlist": "PlayerListPanel",
            "gamelist": "GameListPanel",
            "archivelist": "ArchiveListPanel",
        }
        return x[name] if name in x else name



#############
# ASEAN-Chess
#############

from pychess.Utils.const import (
    VARIANTS_ASEAN,
    ASEANCHESS,
    MAKRUKCHESS,
    QUEEN_PROMOTION,
    CAMBODIANCHESS,
    AIWOKCHESS,
    SITTUYINCHESS,
    NORMAL_MOVE,
    A1,
    A3,
    A6,
    A8,
    B2,
    B3,
    B6,
    B7,
    C3,
    C6,
    D3,
    D4,
    D5,
    D6,
    E3,
    E4,
    E5,
    E6,
    F3,
    F6,
    G2,
    G3,
    G6,
    G7,
    H1,
    H3,
    H6,
    H8,
)

from pychess.Utils.Board import Board

ASEANSTART = "rnbqkbnr/8/pppppppp/8/8/PPPPPPPP/8/RNBQKBNR w - - 0 1"


class AseanBoard(Board):
    variant = ASEANCHESS
    __desc__ = _(
        "ASEAN: http://www.ncf-phil.org/downloadables/2014/May/Asean_chess/Laws_of_ASEAN_Chess_2011_Nov_26.doc"
    )
    name = _("ASEAN")
    cecp_name = "asean"
    need_initial_board = True
    standard_rules = False
    variant_group = VARIANTS_ASEAN

    def __init__(self, setup=False, lboard=None):
        if setup is True:
            Board.__init__(self, setup=ASEANSTART, lboard=lboard)
        else:
            Board.__init__(self, setup=setup, lboard=lboard)


#######################
# Makruk, or Thai chess
#######################
MAKRUKSTART = "rnsmksnr/8/pppppppp/8/8/PPPPPPPP/8/RNSKMSNR w - - 0 1"


class MakrukBoard(Board):
    variant = MAKRUKCHESS
    __desc__ = _("Makruk: http://en.wikipedia.org/wiki/Makruk")
    name = _("Makruk")
    cecp_name = "makruk"
    need_initial_board = True
    standard_rules = False
    variant_group = VARIANTS_ASEAN

    PROMOTION_ZONE = (
        (A6, B6, C6, D6, E6, F6, G6, H6),
        (A3, B3, C3, D3, E3, F3, G3, H3),
    )
    PROMOTIONS = (QUEEN_PROMOTION,)

    def __init__(self, setup=False, lboard=None):
        if setup is True:
            Board.__init__(self, setup=MAKRUKSTART, lboard=lboard)
        else:
            Board.__init__(self, setup=setup, lboard=lboard)


##################################
# Ouk Chatrang, or Cambodian Makruk
##################################
# DEde in cambodian starting fen indicate
# that queens and kings are virgins (not moved yet)
KAMBODIANSTART = "rnsmksnr/8/pppppppp/8/8/PPPPPPPP/8/RNSKMSNR w DEde - 0 1"


class CambodianBoard(Board):
    variant = CAMBODIANCHESS
    __desc__ = _("Cambodian: http://www.khmerinstitute.org/culture/ok.html")
    name = _("Cambodian")
    cecp_name = "cambodian"
    need_initial_board = True
    standard_rules = False
    variant_group = VARIANTS_ASEAN

    PROMOTION_ZONE = (
        (A6, B6, C6, D6, E6, F6, G6, H6),
        (A3, B3, C3, D3, E3, F3, G3, H3),
    )
    PROMOTIONS = (QUEEN_PROMOTION,)

    def __init__(self, setup=False, lboard=None):
        if setup is True:
            Board.__init__(self, setup=KAMBODIANSTART, lboard=lboard)
        else:
            Board.__init__(self, setup=setup, lboard=lboard)


###############
# Ai-Wok Makruk
###############

AIWOKSTART = "rnsaksnr/8/pppppppp/8/8/PPPPPPPP/8/RNSKASNR w - - 0 1"


class AiWokBoard(Board):
    variant = AIWOKCHESS
    __desc__ = _(
        "Ai-Wok: http://www.open-aurec.com/wbforum/viewtopic.php?p=199364&sid=20963a1de2c164050de019e5ed6bf7c4#p199364"
    )
    name = _("Ai-Wok")
    cecp_name = "ai-wok"
    need_initial_board = True
    standard_rules = False
    variant_group = VARIANTS_ASEAN

    PROMOTION_ZONE = (
        (A6, B6, C6, D6, E6, F6, G6, H6),
        (A3, B3, C3, D3, E3, F3, G3, H3),
    )
    PROMOTIONS = (QUEEN_PROMOTION,)

    def __init__(self, setup=False, lboard=None):
        if setup is True:
            Board.__init__(self, setup=AIWOKSTART, lboard=lboard)
        else:
            Board.__init__(self, setup=setup, lboard=lboard)


############################
# Sittuyin, or Myanmar Chess
############################
# Official rules:
# https://vdocuments.net/how-to-play-myanmar-traditional-chess-eng-book-1.html

SITTUYINSTART = "8/8/4pppp/pppp4/4PPPP/PPPP4/8/8/rrnnssfkRRNNSSFK w - - 0 1"


class SittuyinBoard(Board):
    variant = SITTUYINCHESS
    __desc__ = _("Sittuyin: http://en.wikipedia.org/wiki/Sittuyin")
    name = _("Sittuyin")
    cecp_name = "sittuyin"
    need_initial_board = True
    standard_rules = False
    variant_group = VARIANTS_ASEAN

    PROMOTION_ZONE = (
        (A8, B7, C6, D5, E5, F6, G7, H8),
        (A1, B2, C3, D4, E4, F3, G2, H1),
    )
    PROMOTIONS = (QUEEN_PROMOTION, NORMAL_MOVE)

    def __init__(self, setup=False, lboard=None):
        if setup is True:
            Board.__init__(self, setup=SITTUYINSTART, lboard=lboard)
        else:
            Board.__init__(self, setup=setup, lboard=lboard)
